/*------------------------------------------------------------------------------
 bin/djv_info/app.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "app.h"
#include "etc.h"
#include <djv_image/app.h>
#include <djv_image/io.h>
#include <djv_base/dir.h>
#include <djv_base/sys.h>

namespace djv_info {

App::App(int argc, char ** argv) :
  djv_image::app::App("djv_info", argc, argv),
  _verbose(false),
  _seq(file::SEQ_RANGE),
  _recurse(false),
  _columns(sys::terminal_width())
{
  DJV_DBG_OBJ("App::App");
    
  // Command line.

  try
  {
    Str tmp = cmdln_arg();
    cmdln(tmp);
  }
  catch (Error in)
  {
    error(app::error_cmdln, in);
    exit(app::EXIT_ERROR);
  }
  if (exit() != app::EXIT_DEFAULT) return;

  // Setup.
  
  List<File> list;
  for (uint i = 0; i < _input.size(); ++i)
  {
    File file(_input[i]);
    if (_seq)
    {
      if (file.seq_wildcard())
        file = file::seq_match(file, Dir(file.path).list());
      if (file.seq_valid()) file.type = file::SEQ;
    }
    
    List<Str> tmp = file::seq_expand(file);
    for (uint j = 0; j < tmp.size(); ++j)
    {
      file = File(tmp[j]);
      if (_seq && file.seq_valid()) file.type = file::SEQ;
      list += file;
    }
  }

  file::seq_compress(&list, _seq);
  
  DJV_DBG_PRN("files = " << list);
  
  // Work.
  
  for (uint i = 0; i < list.size(); ++i)
    if (file::DIR == list[i].type)
    {
      try { dir(list[i], (list.size() > 1) || _recurse); }
      catch (Error error)
      {
        this->error(error, list[i]);
        exit(app::EXIT_ERROR);
      }
    }
    else
    {
      try { file(list[i]); }
      catch (Error error)
      {
        this->error(error, list[i].name());
        exit(app::EXIT_ERROR);
      }
    }
}

void App::file(const File & in, bool path) throw (Error)
{
  DJV_DBG_OBJ("App::file");
  DJV_DBG_PRN("in = " << in);
  
  // Open.

  List<djv_image::Info> info;
  djv_image::Tag tag;
  djv_image::Time time;
  std::auto_ptr<djv_image::io::Load>(
    djv_image::io::global()->load(in, &info, &tag, &time));
  
  // Print.

  Str name = in.get(-1, path);
  
  if (! _verbose)
  {
    const Str str = str::Format(label_info).
      arg(info[0].size.x).arg(info[0].size.y).arg(aspect(info[0].size), 0, -2).
      arg(djv_image::pixel::label_pixel()[info[0].pixel]).
      arg(djv_image::time::units_to_str(time.list.size(), time.speed)).
      arg(djv_image::time::speed_to_float(time.speed), 0, -2);
    print(str::Format("%% %%").
      arg(name).
      arg(str, _columns - int(name.size()) - 2));
    
    for (uint i = 1; i < info.size(); ++i)
    {
      const Str name_str = str::Format(label_layer).arg(i).arg(info[i].name);
      const Str info_str = str::Format(label_layer_info).
        arg(info[i].size.x).arg(info[i].size.y).
        arg(aspect(info[i].size), 0, -2).
        arg(djv_image::pixel::label_pixel()[info[i].pixel]);
      print(str::Format("%% %%").
        arg(name_str).
        arg(info_str, _columns - int(name_str.size()) - 2));
    }
  }
  else
  {
    print(name);

    print(str::Format(label_width).arg(info[0].size.x));
    print(str::Format(label_height).arg(info[0].size.y));
    print(str::Format(label_aspect).arg(aspect(info[0].size), 0, -2));
    print(str::Format(label_pixel).
      arg(djv_image::pixel::label_pixel()[info[0].pixel]));
    
    for (uint i = 1; i < info.size(); ++i)
    {
      print(str::Format(label_layer).arg(info[i].name));
      print(str::Format(label_width).arg(info[i].size.x));
      print(str::Format(label_height).arg(info[i].size.y));
      print(str::Format(label_aspect).arg(aspect(info[i].size), 0, -2));
      print(str::Format(label_pixel).
        arg(djv_image::pixel::label_pixel()[info[i].pixel]));
    }

    print(str::Format(label_start).
      arg(djv_image::time::units_to_str(time.start(), time.speed)));
    print(str::Format(label_end).
      arg(djv_image::time::units_to_str(time.end(), time.speed)));
    print(str::Format(label_duration).
      arg(djv_image::time::units_to_str(time.list.size(), time.speed)));
    print(str::Format(label_speed).
      arg(djv_image::time::speed_to_float(time.speed)));

    const List<Str> key_list = tag.key_list();
    for (uint i = 0; i < key_list.size(); ++i)
      print(str::Format(label_tag).arg(key_list[i]).arg(tag[key_list[i]]));

    separator();
  }
}

void App::dir(const File & in, bool label) throw (Error)
{
  DJV_DBG_OBJ("App::dir");
  DJV_DBG_PRN("in = " << in);
  
  // Setup.

  Dir dir(in);
  if (! dir.valid()) throw Error(error_dir);
  
  List<File> list = dir.list();
  file::filter(&list, file::FILTER_DIR);
  file::seq_compress(&list, _seq);

  // Print.

  if (label)
  {
    print(str::Format(label_dir).arg(in));
    if (_verbose) separator();
  }
  
  for (uint i = 0; i < list.size(); ++i)
  {
    try { file(list[i]); }
    catch (Error) {}
  }
  
  separator();
  
  // Recurse.
  
  if (_recurse)
  {
    List<File> list = Dir(in).list();
    file::filter(&list, file::FILTER_FILE | file::FILTER_HIDDEN);
    for (uint i = 0; i < list.size(); ++i)
      this->dir(list[i], label);
  }
}

}

int main(int argc, char ** argv)
{
  int r = 1;
  try { r = djv_info::App(argc, argv).exec(); }
  catch (djv_base::Error in) { djv_base::error::print(in); }
  return r;
}

