/*------------------------------------------------------------------------------
 bin/djv_view/image_group.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include "image.h"
#include "image_display.h"
#include "image_histogram.h"
#include "image_info.h"
#include "image_magnify.h"
#include "image_sample.h"
#include "shortcut.h"
#include <djv_ui/app.h>
#include <djv_ui/button.h>
#include <djv_ui/choice.h>
#include <djv_ui/label.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/style.h>
#include <djv_image/io.h>

namespace djv_view {
namespace image {

struct Group::_Group
{  
  _Group() :
    menu(0)
  {}

  djv_ui::menu::Widget * menu;
  List<int> menu_scale, menu_rotate, menu_profile, menu_channel, menu_display;
  djv_ui::button::Tool * display, * magnify, * sample;
  djv_ui::label::Widget * pick_label;
};

Group::Group(
  djv_ui::menu::Widget * menu,
  djv_ui::layout::Horizontal * toolbar,
  const Group * copy
) :
  redraw_sig(this),
  resize_sig(this),
  frame_store_sig(this),
  display_sig(this),
  display_value_sig(this),
  magnify_sig(this),
  sample_sig(this),
  histogram_sig(this),
  info_sig(this),
  update_sig(this),
  _frame_store(copy ? copy->_frame_store : false),
  _mirror(copy ? copy->_mirror : prefs()->mirror()),
  _scale(copy ? copy->_scale : prefs()->scale()),
  _rotate(copy ? copy->_rotate : prefs()->rotate()),
  _profile(copy ? copy->_profile : prefs()->profile()),
  _channel(copy ? copy->_channel : prefs()->channel()),
  _display(copy ? copy->_display : prefs()->display_value()),
  _p(new _Group)
{
  using namespace djv_ui;
  
  DJV_DBG_OBJ("Group::Group");
  DJV_DBG_PRN("color = " << _display.color);
  
  // Widgets.
  
  _p->menu = menu;
  
  _p->display = new button::Tool("image_display");
  _p->magnify = new button::Tool("magnify");
  _p->sample = new button::Tool("color_sample");
  
  _p->pick_label = new label::Widget;
  _p->pick_label->size_str(toolbar_pick_format);
  _p->pick_label->tooltip(tooltip_pick);
  
  // Layout.
  
  toolbar->add(_p->display);
  toolbar->add(_p->magnify);
  toolbar->add(_p->sample);
  toolbar->add_spacer(style::global()->spacing());
  toolbar->add(_p->pick_label);
  toolbar->add_spacer(style::global()->spacing());
  
  // Initialize.
  
  // Callbacks.
  
  _p->display->sig.set(this, display_callback);
  _p->magnify->sig.set(this, magnify_callback);
  _p->sample->sig.set(this, sample_callback);
  
  prefs()->mirror_sig.set(this, mirror_callback);
  prefs()->scale_sig.set(this, scale_callback);
  prefs()->rotate_sig.set(this, rotate_callback);
  prefs()->profile_sig.set(this, profile_callback);
  prefs()->channel_sig.set(this, channel_callback);
  prefs()->display_value_sig.set(this, display_callback);
  prefs()->display_list_sig.set(this, display_list_callback);
}

Group::~Group()
{}

void Group::pick(const Vec2i & in, double zoom)
{
  const Vec2i tmp = vec::floor<double, int>(Vec2f(in) / zoom);
  _p->pick_label->set(str::Format(toolbar_pick).arg(tmp.x).arg(tmp.y));
}

void Group::display(const Display & in) { _display = in; }

void Group::menu_update(djv_ui::menu::Group * in)
{
  using namespace djv_ui::menu;
  
  const List<djv_ui::shortcut::Shortcut> & shortcut =
    shortcut::prefs()->shortcut();
  
  in->add(menu_title, 0, 0, 0, SUB_MENU);

  in->add(
    menu_frame_store, shortcut[shortcut::IMAGE_FRAME_STORE].value,
    _frame_store_callback, this,
    TOGGLE, _frame_store);

  in->add(
    menu_frame_store_load, shortcut[shortcut::IMAGE_FRAME_STORE_LOAD].value,
    _frame_store_load_callback, this,
    DIVIDER);

  in->add(
    menu_mirror_h, shortcut[shortcut::IMAGE_MIRROR_HORIZONTAL].value,
    _mirror_h_callback, this,
    TOGGLE, _mirror.x);

  in->add(
    menu_mirror_v, shortcut[shortcut::IMAGE_MIRROR_VERTICAL].value,
    _mirror_v_callback, this,
    TOGGLE, _mirror.y);
  
  in->add(menu_scale, 0, 0, 0, SUB_MENU);
  _p->menu_scale = in->add(
    label_scale(), List<int>() <<
      shortcut[shortcut::IMAGE_SCALE_NONE].value <<
      shortcut[shortcut::IMAGE_SCALE_16_9].value <<
      shortcut[shortcut::IMAGE_SCALE_1_85].value <<
      shortcut[shortcut::IMAGE_SCALE_2_35].value,
    _scale_callback, this,
    RADIO, _scale);
  in->end();
  
  in->add(menu_rotate, 0, 0, 0, SUB_MENU | DIVIDER);
  _p->menu_rotate = in->add(
    label_rotate(), List<int>() <<
      shortcut[shortcut::IMAGE_ROTATE_0].value <<
      shortcut[shortcut::IMAGE_ROTATE_90].value <<
      shortcut[shortcut::IMAGE_ROTATE_180].value <<
      shortcut[shortcut::IMAGE_ROTATE_270].value,
    _rotate_callback, this,
    RADIO, _rotate);
  in->end();

  in->add(
    menu_profile, shortcut[shortcut::IMAGE_PROFILE].value,
    _profile_callback, this,
    TOGGLE, _profile);
  
  in->add(menu_channel, 0, 0, 0, SUB_MENU);
  _p->menu_channel = in->add(
    label_channel(), List<int>() <<
      shortcut[shortcut::IMAGE_CHANNEL_DEFAULT].value <<
      shortcut[shortcut::IMAGE_CHANNEL_1].value <<
      shortcut[shortcut::IMAGE_CHANNEL_2].value <<
      shortcut[shortcut::IMAGE_CHANNEL_3].value <<
      shortcut[shortcut::IMAGE_CHANNEL_4].value,
    _channel_callback, this,
    RADIO, _channel);
  in->end();
  
  in->add(menu_display, 0, 0, 0, SUB_MENU | DIVIDER);
  
  in->add(
    menu_display_show, shortcut[shortcut::IMAGE_DISPLAY].value,
    _display_callback, this);
  
  in->add(
    menu_display_reset, shortcut[shortcut::IMAGE_DISPLAY_RESET].value,
    _display_reset_callback, this,
    DIVIDER);
  
  in->add(menu_display_add, 0, _display_add_callback, this);
  in->add(menu_display_del, 0, _display_del_callback, this);
  in->add(menu_display_del_all, 0, _display_del_all_callback, this, DIVIDER);
  
  _p->menu_display = in->add(
    prefs()->display_name_list(), List<int>() <<
      shortcut[shortcut::IMAGE_DISPLAY_1].value <<
      shortcut[shortcut::IMAGE_DISPLAY_2].value <<
      shortcut[shortcut::IMAGE_DISPLAY_3].value <<
      shortcut[shortcut::IMAGE_DISPLAY_4].value <<
      shortcut[shortcut::IMAGE_DISPLAY_5].value <<
      shortcut[shortcut::IMAGE_DISPLAY_6].value <<
      shortcut[shortcut::IMAGE_DISPLAY_7].value <<
      shortcut[shortcut::IMAGE_DISPLAY_8].value <<
      shortcut[shortcut::IMAGE_DISPLAY_9].value <<
      shortcut[shortcut::IMAGE_DISPLAY_10].value,
    _display_list_callback, this);
  in->end();
  
  in->add(
    menu_magnify, shortcut[shortcut::IMAGE_MAGNIFY].value,
    _magnify_callback, this);
  
  in->add(
    menu_sample, shortcut[shortcut::IMAGE_SAMPLE].value,
    _sample_callback, this);
  
  in->add(
    menu_histogram, shortcut[shortcut::IMAGE_HISTOGRAM].value,
    _histogram_callback, this);
  
  in->add(
    menu_info, shortcut[shortcut::IMAGE_INFO].value,
    _info_callback, this);
  
  in->end();
}

void Group::toolbar_update()
{
  using djv_ui::shortcut::serialize;
  
  const List<djv_ui::shortcut::Shortcut> & shortcut =
    shortcut::prefs()->shortcut();
  
  _p->display->tooltip(str::Format(tooltip_display).
    arg(serialize(shortcut[shortcut::IMAGE_DISPLAY].value)));
  _p->magnify->tooltip(str::Format(tooltip_magnify).
    arg(serialize(shortcut[shortcut::IMAGE_MAGNIFY].value)));
  _p->sample->tooltip(str::Format(tooltip_sample).
    arg(serialize(shortcut[shortcut::IMAGE_SAMPLE].value)));
}

bool Group::frame_store() const { return _frame_store; }

void Group::frame_store_callback(bool in)
{
  DJV_DBG_OBJ("Group::frame_store_callback");
  DJV_DBG_PRN("in = " << in);
  
  _frame_store = in;
  
  redraw_sig.emit(true);
  update_sig.emit(true);
}

void Group::frame_store_load_callback(bool)
{
  DJV_DBG_OBJ("Group::frame_store_load_callback");
  
  frame_store_sig.emit(true);
}

void Group::_frame_store_callback()
{
  frame_store_callback(_p->menu->value());
}

void Group::_frame_store_load_callback()
{
  frame_store_load_callback(true);
}

const Vec2b & Group::mirror() const { return _mirror; }

void Group::mirror_callback(const Vec2b & in)
{
  DJV_DBG_OBJ("Group::mirror_callback");
  DJV_DBG_PRN("in = " << in);
  
  _mirror = in;
  
  redraw_sig.emit(true);
  update_sig.emit(true);
}

void Group::_mirror_h_callback()
{
  mirror_callback(Vec2b(_p->menu->value(), _mirror.y));
}

void Group::_mirror_v_callback()
{
  mirror_callback(Vec2b(_mirror.x, _p->menu->value()));
}

SCALE Group::scale() const { return _scale; }

void Group::scale_callback(SCALE in)
{
  DJV_DBG_OBJ("Group::scale_callback");
  //DJV_DBG_PRN("in = " << in);
  
  _scale = in;
  
  redraw_sig.emit(true);
  resize_sig.emit(true);
  update_sig.emit(true);
}

void Group::_scale_callback()
{
  scale_callback(SCALE(list::find(_p->menu->item(), _p->menu_scale)));
}

ROTATE Group::rotate() const { return _rotate; }

void Group::rotate_callback(ROTATE in)
{
  DJV_DBG_OBJ("Group::rotate_callback");
  //DJV_DBG_PRN("in = " << in);
  
  _rotate = in;
  
  redraw_sig.emit(true);
  resize_sig.emit(true);
  update_sig.emit(true);
}

void Group::_rotate_callback()
{
  rotate_callback(ROTATE(list::find(_p->menu->item(), _p->menu_rotate)));
}

bool Group::profile() const { return _profile; }

void Group::profile_callback(bool in)
{
  DJV_DBG_OBJ("Group::profile_callback");
  DJV_DBG_PRN("in = " << in);
  
  _profile = in;
  
  redraw_sig.emit(true);
  update_sig.emit(true);
}

void Group::_profile_callback() { profile_callback(_p->menu->value()); }

CHANNEL Group::channel() const { return _channel; }

void Group::channel_callback(CHANNEL in)
{
  DJV_DBG_OBJ("Group::channel_callback");
  DJV_DBG_PRN("in = " << in);

  _channel = in;
  
  redraw_sig.emit(true);
  update_sig.emit(true);
}

void Group::_channel_callback()
{
  channel_callback(CHANNEL(list::find(_p->menu->item(), _p->menu_channel)));
}

const Display & Group::display() const { return _display; }

void Group::display_callback(bool) { display_sig.emit(true); }
void Group::display_callback(const Display & in) { display_value_sig.emit(in); }
void Group::_display_callback() { display_callback(true); }

void Group::display_reset_callback(bool) { display_callback(Display()); }
void Group::_display_reset_callback() { display_reset_callback(true); }

void Group::display_add_callback(const Str & in)
{
  DJV_DBG_OBJ("Group::display_add_callback");
  DJV_DBG_PRN("in = " << in);
  
  List<Display> list = prefs()->display_list();
  Display tmp = _display;
  tmp.name = in;
  list += tmp;
  prefs()->display_list(list);
}

void Group::_display_add_callback()
{
  djv_ui::dialog::global()->input(
    dialog_display_add,
    "",
    this,
    display_add_callback);
}

void Group::display_del_callback(int in)
{
  DJV_DBG_OBJ("Group::display_del_callback");
  DJV_DBG_PRN("in = " << in);
  
  List<Display> list = prefs()->display_list();
  List<Display>::iterator i = list.begin();
  for (int j = 0; j < in && i != list.end(); ++j, ++i) ;
  list.erase(i);
  prefs()->display_list(list);
}

void Group::_display_del_callback()
{
  const List<Str> list = prefs()->display_name_list();
  if (! list.size()) return;
  
  djv_ui::dialog::global()->choice(
    dialog_display_del,
    list,
    this,
    display_del_callback);
}

void Group::display_del_all_callback(bool)
{
  prefs()->display_list(List<Display>());
}

void Group::_display_del_all_callback()
{
  const List<Str> list = prefs()->display_name_list();
  if (! list.size()) return;
  
  djv_ui::dialog::global()->question(
    dialog_display_del_all,
    this,
    display_del_all_callback);
}

void Group::display_list_callback(bool)
{
  DJV_DBG_OBJ("Group::display_list_callback");
  
  update_sig.emit(true);
}

void Group::_display_list_callback()
{
  display_callback(
    prefs()->display_list()[list::find(_p->menu->item(), _p->menu_display)]);
}

void Group::magnify_callback(bool) { magnify_sig.emit(true); }
void Group::_magnify_callback() { magnify_callback(true); }

void Group::sample_callback(bool) { sample_sig.emit(true); }
void Group::_sample_callback() { sample_callback(true); }

void Group::_histogram_callback() { histogram_sig.emit(true); }

void Group::_info_callback() { info_sig.emit(true); }

}
}

