/*------------------------------------------------------------------------------
 bin/djv_view/image_sample.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include "app.h"
#include "image_sample.h"
#include "view.h"
#include <djv_ui/button.h>
#include <djv_ui/choice.h>
#include <djv_ui/color.h>
#include <djv_ui/group.h>
#include <djv_ui/label.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/num_int.h>
#include <djv_ui/prefs.h>

namespace djv_view {
namespace image_sample {

struct Dialog::_Dialog
{
  std::auto_ptr<djv_image::gl_offscreen::Buffer> buffer;
  std::auto_ptr<djv_image::gl_data::Data> data;

  djv_ui::color::Widget * widget;
  djv_ui::color::Swatch * swatch;
  djv_ui::choice::Widget * size;
  djv_ui::button::Check * profile;
  djv_ui::button::Check * lock;
  djv_ui::button::Push * close;
};

Dialog::Dialog() :
  Base(dialog_title),
  _view(0),
  _value(djv_image::pixel::RGBA_F32),
  _size(SIZE_3),
  _profile(true),
  _lock(false),
  _p(new _Dialog)
{
  using namespace djv_ui;
  
  DJV_DBG_OBJ("Dialog::Dialog");
  
  // Widgets.

  _p->widget = new color::Widget;
  _p->swatch = new color::Swatch;
  _p->swatch->size(layout::STRETCH, layout::STRETCH);
  
  _p->size = new choice::Widget(label_size());
  label::Widget * size_label = new label::Widget(dialog_size);
  _p->size->tooltip(tooltip_size);
  
  _p->profile = new button::Check(dialog_profile);
  _p->profile->tooltip(tooltip_profile);
  
  _p->lock = new button::Check(dialog_lock);
  _p->lock->tooltip(tooltip_lock);

  _p->close = new button::Push(djv_ui::dialog::label_close);
  
  // Layout.
  
  Layout_V * layout = new Layout_V(this);
  
  Layout_H * layout_h = new Layout_H(layout);
  layout_h->margin(0);
  layout_h->add(_p->swatch);
  layout_h->add(_p->widget);
  layout_h->stretch(_p->widget);
  
  layout->add_spacer(-1, true);
  
  layout_h = new Layout_H(layout);
  layout_h->margin(0);
  layout_h->add(size_label);
  layout_h->add(_p->size);
  layout_h->add(_p->profile);
  layout_h->add(_p->lock);
  layout_h->add_spacer(-1, true);
  layout_h->add(_p->close);
  
  // Preferences.
  
  prefs::Prefs prefs(prefs::prefs(), "image_sample");
  prefs::get(&prefs, "size", &_size);
  djv_image::PIXEL pixel = _value.pixel();
  prefs::get(&prefs, "pixel", &pixel);
  _value.set(pixel);
  prefs::get(&prefs, "profile", &_profile);
  prefs::get(&prefs, "lock", &_lock);
  
  // Initialize.

  widget_update();
  
  size(size_hint());
  
  // Callbacks.

  _p->widget->sig.set(this, widget_callback);
  _p->size->sig.set(this, size_callback);
  _p->profile->sig.set(this, profile_callback);
  _p->lock->sig.set(this, lock_callback);
  _p->close->sig.set(this, close_callback);
}

Dialog::~Dialog()
{
  DJV_DBG_OBJ("Dialog::~Dialog");
}

void Dialog::del()
{
  using namespace djv_ui;
  
  DJV_DBG_OBJ("Dialog::del");
  
  djv_ui::dialog::Base::del();

  prefs::Prefs prefs(prefs::prefs(), "image_sample");
  prefs::set(&prefs, "size", _size);
  prefs::set(&prefs, "pixel", _value.pixel());
  prefs::set(&prefs, "profile", _profile);
  prefs::set(&prefs, "lock", _lock);
}

void Dialog::set(const view::Widget * in, const Vec2i & pick)
{
  _view = in;
  _pick = pick;
  
  widget_update();
}

void Dialog::pick(const view::Widget * in, const Vec2i & pick)
{
  if (! shown() || ! visible()) return;
  
  set(in, pick);
}

void Dialog::update(const view::Widget * in)
{
  if (in != _view || ! shown() || ! visible()) return;
  
  widget_update();
}

void Dialog::del(const view::Widget * in)
{
  if (in != _view) return;
  
  _view = 0;
}

void Dialog::show()
{
  _p->close->take_focus();

  djv_ui::dialog::Base::show();
}

void Dialog::widget_callback(const djv_image::Color & in)
{
  _value = in;
  
  _p->swatch->set(in);
}

void Dialog::size_callback(int in)
{
  _size = SIZE(in);
  
  widget_update();
}

void Dialog::profile_callback(bool in)
{
  _profile = in;
  
  widget_update();
}

void Dialog::lock_callback(bool in)
{
  _lock = in;
  
  widget_update();
}

void Dialog::close_callback(bool)
{
  hide();
}

void Dialog::widget_update()
{
  if (! _view) return;
  
  DJV_DBG_OBJ("Dialog::widget_update");
  DJV_DBG_PRN("profile = " << _profile);
  DJV_DBG_PRN("lock = " << _lock);
  
  callbacks(false);

  const djv_image::Data * data = _view->get();
  if (! _lock && data) _value.set(data->pixel());
  else _value.set(_p->widget->get().pixel());
  
  // Pick.
  
  const Vec2i pick = vec::floor<double, int>(Vec2f(_pick) / _view->zoom());
  DJV_DBG_PRN("pick = " << pick);
  
  // Render color sample.
  
  const int size = image_sample::size(_size);
  DJV_DBG_PRN("size = " << size);

  djv_image::Data tmp(djv_image::Info(Vec2i(size), _value.pixel()));
  try
  {
    if (! _p->buffer.get())
      _p->buffer = std::auto_ptr<djv_image::gl_offscreen::Buffer>(
        djv_image::gl_offscreen::global()->buffer_create());
    _p->buffer->init(tmp.info());

    if (! _p->data.get())
      _p->data = std::auto_ptr<djv_image::gl_data::Data>(
        new djv_image::gl_data::Data);

    djv_image::gl_data::Option option = _view->option();
    option.xform.position -= pick - Vec2i((size - 1) / 2);
    if (! _profile) option.profile = djv_image::Profile();

    djv_image::Data empty;
    if (! data) data = &empty;
    djv_image::gl_data::copy(
      *data, &tmp, option, _p->data.get(), _p->buffer.get());

    djv_image::gl_data::average(tmp, &_value);
  }
  catch (Error in)
  {
    app()->error(in);
  }
  DJV_DBG_PRN("value = " << _value);
  
  _p->widget->set(_value);
  _p->swatch->set(_value);
  _p->size->set(_size);
  _p->profile->set(_profile);
  _p->lock->set(_lock);
  
  callbacks(true);
}

int size(SIZE in)
{
  static const int data [] =
  {
    1,
    3,
    5,
    7,
    9,
    51,
    101
  };
  return data[in];
}

DJV_STR_OPERATOR_SET(SIZE, label_size())
DJV_STR_OPERATOR_GET(SIZE, label_size())

const List<Str> & label_size()
{
  static const List<Str> data = List<Str>() <<
    "1x1" <<
    "3x3" <<
    "5x5" <<
    "7x7" <<
    "9x9" <<
    "51x51" <<
    "101x101";
  return data;
}

const Str
  dialog_title = "Image Sample",
  dialog_size = "Size:",
  dialog_profile = "Profile",
  dialog_lock = "Lock";

const Str
  tooltip_size = "Color sample size.",
  tooltip_profile = "Use image color profile.",
  tooltip_lock = "Lock pixel.";

}
}

