/*------------------------------------------------------------------------------
 bin/djv_view/widget.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include "widget.h"
#include <djv_ui/image.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/prefs.h>
#include <djv_image/time.h>
#include <FL/fl_draw.H>

namespace djv_view {
namespace widget {

Frame_Widget::Frame_Widget() :
  sig(this),
  _value(0)
{
  // Widgets.
  
  _widget = new djv_ui::text::Edit;
  
  djv_ui::num::Inc * inc_widget = new djv_ui::num::Inc;
  
  // Layout.
  
  Item::size(djv_ui::layout::FIXED, djv_ui::layout::FIXED);
  Item::align(djv_ui::layout::LEFT, djv_ui::layout::CENTER);
  
  djv_ui::Layout_H * layout = new djv_ui::Layout_H(this);
  layout->spacing(2);
  layout->margin(0);
  layout->add(_widget);
  layout->stretch(_widget);
  layout->add(inc_widget);
  
  // Initialize.
  
  widget_update();
  
  // Callbacks.
  
  _widget->sig.set(this, widget_callback);
  inc_widget->inc_sig.set(this, inc_callback);
  inc_widget->dec_sig.set(this, dec_callback);
}

void Frame_Widget::set(int in)
{
  if (in == _value) return;
  
  _value = in;
  
  widget_update();
  
  sig.emit(_value);
}

int Frame_Widget::get() const { return _value; }

void Frame_Widget::list(const List<int> & in)
{
  if (in == _list) return;
  
  _list = in;
  
  widget_update();
}

const List<int> & Frame_Widget::list() const { return _list; }

void Frame_Widget::speed(const djv_image::time::Speed & in)
{
  if (in == _speed) return;
  
  _speed = in;
  
  widget_update();
}

const djv_image::time::Speed & Frame_Widget::speed() const { return _speed; }

void Frame_Widget::text_font(const Fl_Font & in)
{
  Widget::text_font(in);
  _widget->text_font(in);
}

const Fl_Font & Frame_Widget::text_font() const
{
  return Widget::text_font();
}

void Frame_Widget::text_color(const Fl_Color & in)
{
  Widget::text_color(in);
  _widget->text_color(in);
}

const Fl_Color & Frame_Widget::text_color() const
{
  return Widget::text_color();
}

void Frame_Widget::dirty()
{
  Str size_str;
  switch (djv_image::time::global()->units())
  {
    case djv_image::time::UNITS_TIMECODE:
      size_str = label_timecode_format;
      break;

    case djv_image::time::UNITS_FRAMES:
      size_str = label_frame_format;
      break;
  }
  _widget->size_str(size_str);
    
  Widget::dirty();
  
  widget_update();
}

djv_ui::layout::Item * Frame_Widget::group_align() { return _widget; }

void Frame_Widget::widget_callback(const Str & in)
{
  set(file::seq_find(djv_image::time::str_to_units(in, _speed), _list));
  
  widget_update();
}

void Frame_Widget::inc_callback(bool) { set(_value + 1); }
void Frame_Widget::dec_callback(bool) { set(_value - 1); }

void Frame_Widget::widget_update()
{
  callbacks(false);
  
  _widget->set(djv_image::time::units_to_str(_list.size() ? _list[_value] : 0,
    _speed));
  
  callbacks(true);
}

Frame_Slider::Frame_Slider() :
  _in(0),
  _out(0)
{
  //frame(FL_FLAT_BOX);
}

void Frame_Slider::in_out(int in, int out)
{
  _in = in;
  _out = out;
  
  redraw();
}

int Frame_Slider::in() const { return _in; }
int Frame_Slider::out() const { return _out; }
  
void Frame_Slider::min_action() { set(_in > min() ? _in : min()); }
void Frame_Slider::max_action() { set(_out < max() ? _out : max()); }

void Frame_Slider::cache(const List<int> & in)
{
  _cache = in;
  
  redraw();
}

int Frame_Slider::handle(int in)
{
  switch (in)
  {
    case FL_KEYBOARD: return 0;
  }
  return djv_ui::slider::Int::handle(in);
}

void Frame_Slider::draw()
{
  DJV_DBG_OBJ("Frame_Slider::draw");
  
  Frame::draw();

  const Box2i & box = slider_geom();
  const int range = max() - min();

  if (_in > min() || _out < max() && min() != max())
  {
    const double v0 = (math::max(_in, min()) - min()) / double(range);
    const double v1 = (math::min(_out, max()) - min()) / double(range);

    fl_color(FL_SELECTION_COLOR);
    fl_rectf(
      box.x + math::floor(box.w * v0),
      box.y,
      math::floor(box.w * (v1 - v0)) + 1,
      box.h);
  }
  
  /*if (math::floor(1 / double(range) * box.w) > 1)
  {
    fl_color(fl_darker(FL_BACKGROUND2_COLOR));
    for (int i = 0; i <= range; ++i)
    {
      const double v = i / double(range);
    
      fl_rectf(box.x + math::floor(box.w * v), box.y, 1, box.h);
    }
  }*/
  
  if (_cache.size())
  {
    fl_color(FL_BLUE);
    
    const List<math::Range<int> > list = math::range(_cache);
    for (uint i = 0; i < list.size(); ++i)
    {
      const double v0 = (list[i].min - min()) / double(range);
      const double v1 = (list[i].max - min()) / double(range);

      fl_rectf(
        box.x + math::floor(box.w * v0),
        box.y + box.h / 2,
        math::floor(box.w * (v1 - v0)) + 1,
        box.h / 2);
    }
  }
  
  handle_draw(
    min() != max() ? ((get() - min()) / double(max() - min())) : 0,
    frame_geom());
}

Frame_Display_Widget::Frame_Display_Widget() :
  _value(0)
{
  // Widgets.
  
  _widget = new djv_ui::text::Display;
  
  // Layout.
  
  Item::size(djv_ui::layout::FIXED, djv_ui::layout::FIXED);
  Item::align(djv_ui::layout::LEFT, djv_ui::layout::CENTER);
  
  djv_ui::Layout_H * layout = new djv_ui::Layout_H(this);
  layout->margin(0);
  layout->add(_widget);
  
  // Initialize.

  widget_update();
}

void Frame_Display_Widget::set(int in)
{
  if (in == _value) return;
  
  _value = in;
  
  widget_update();
}

int Frame_Display_Widget::get() const { return _value; }

void Frame_Display_Widget::speed(const djv_image::time::Speed & in)
{
  if (in == _speed) return;
  
  _speed = in;
  
  widget_update();
}

const djv_image::time::Speed & Frame_Display_Widget::speed() const
{
  return _speed;
}

void Frame_Display_Widget::text_font(const Fl_Font & in)
{
  Widget::text_font(in);
  _widget->text_font(in);
}

const Fl_Font & Frame_Display_Widget::text_font() const
{
  return Widget::text_font();
}

void Frame_Display_Widget::text_color(const Fl_Color & in)
{
  Widget::text_color(in);
  _widget->text_color(in);
}

const Fl_Color & Frame_Display_Widget::text_color() const
{
  return Widget::text_color();
}

void Frame_Display_Widget::dirty()
{
  Str size_str;
  switch (djv_ui::image::prefs()->units())
  {
    case djv_image::time::UNITS_TIMECODE:
      size_str = label_timecode_format;
      break;

    case djv_image::time::UNITS_FRAMES:
      size_str = label_frame_format;
      break;
  }
  _widget->size_str(size_str);

  Widget::dirty();

  widget_update();
}

void Frame_Display_Widget::widget_update()
{
  callbacks(false);
  
  _widget->set(djv_image::time::units_to_str(_value, _speed));

  callbacks(true);
}

Speed_Widget::Speed_Widget() :
  sig(this)
{
  _widget = new djv_ui::text::Display;
  _menu = new djv_ui::menu::Popup;

  Item::size(djv_ui::layout::FIXED, djv_ui::layout::FIXED);
  
  djv_ui::Layout_H * layout = new djv_ui::Layout_H(this);
  layout->margin(0);
  layout->add(_widget);
  layout->stretch(_widget);
  layout->add(_menu);

  widget_update();
  menu_update();
  
  _menu->sig.set(this, menu_callback);
}

void Speed_Widget::set(const djv_image::time::Speed & in)
{
  if (in == _value) return;
  
  _value = in;
  
  widget_update();
  sig.emit(_value);
}

const djv_image::time::Speed & Speed_Widget::get() const { return _value; }

void Speed_Widget::set_default(const djv_image::time::Speed & in)
{
  if (in == _value_default) return;
  
  _value_default = in;
  
  menu_update();
}

const djv_image::time::Speed & Speed_Widget::get_default() const
{
  return _value_default;
}

void Speed_Widget::dirty()
{
  _widget->size_str(label_speed_format);

  Widget::dirty();
}

void Speed_Widget::menu_callback(int in)
{
  if (_value_default.valid())
    set(0 == in ? _value_default : djv_image::time::FPS(in - 1));
  else
    set(djv_image::time::FPS(in));
}

void Speed_Widget::widget_update()
{
  callbacks(false);
  
  _widget->set(str::float_to_str(djv_image::time::speed_to_float(_value), -2));
  
  callbacks(true);
}

void Speed_Widget::menu_update()
{
  callbacks(false);
  
  djv_ui::menu::Group group;
  if (_value_default.valid())
    group.add(
      str::float_to_str(djv_image::time::speed_to_float(_value_default), -2),
      0, 0, 0, djv_ui::menu::DIVIDER);
  const List<Str> & label = djv_image::time::label_fps();
  for (uint i = 0; i < label.size(); ++i)
    group.add(label[i]);
  group.end();
  
  _menu->set(group);
  
  callbacks(true);
}

Speed_Display_Widget::Speed_Display_Widget() :
  _value(0.0),
  _dropped_frames(false)
{
  _widget = new djv_ui::text::Display;

  Item::size(djv_ui::layout::FIXED, djv_ui::layout::FIXED);
  
  djv_ui::Layout_H * layout = new djv_ui::Layout_H(this);
  layout->margin(0);
  layout->add(_widget);

  widget_update();
}

void Speed_Display_Widget::set(double in)
{
  if (in == _value) return;
  
  _value = in;
  
  widget_update();
}

void Speed_Display_Widget::dropped_frames(bool in)
{
  if (in == _dropped_frames) return;
  
  _dropped_frames = in;
  
  _widget->text_color(
    _dropped_frames ? FL_SELECTION_COLOR : FL_FOREGROUND_COLOR);
  redraw();
}

void Speed_Display_Widget::dirty()
{
  _widget->size_str(label_speed_format);

  Widget::dirty();
}

void Speed_Display_Widget::widget_update()
{
  callbacks(false);
  
  _widget->set(str::float_to_str(_value, -2));
  
  callbacks(true);
}

const Str
  label_timecode_format = "00:00:00:00",
  label_frame_format = "000000",
  label_speed_format = "000.00";

}
}

