/*------------------------------------------------------------------------------
 bin/djv_view/window.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include "app.h"
#include "image_display.h"
#include "image_histogram.h"
#include "image_info.h"
#include "image_magnify.h"
#include "image_sample.h"
#include "shortcut.h"
#include "window.h"
#include <djv_ui/button.h>
#include <djv_ui/dialog.h>
#include <djv_ui/group.h>
#include <djv_ui/icon.h>
#include <djv_ui/image_io.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/menu.h>
#include <djv_ui/shortcut.h>
#include <djv_ui/style.h>
#include <FL/Fl.H>

namespace djv_view {
namespace window {

struct Widget::_Widget
{
  _Widget() :
    menu(0)
  {}

  djv_ui::menu::Widget * menu;
  djv_ui::group::Frame * toolbar;
  djv_ui::button::Tool * window_new, * window_copy, * window_close,
    * window_fit, * window_fullscreen;
  djv_ui::group::Frame * playbar;
  view::Widget * view;
};

Widget::Widget(const Widget * copy) :
  djv_ui::window::Widget(app()->name()),
  _window_fullscreen(false),
  _window_toolbar(copy ? copy->_window_toolbar : prefs()->toolbar()),
  _window_playbar(copy ? copy->_window_playbar : prefs()->playbar()),
  _image_p(0),
  _image_tmp(copy ? copy->_image_tmp : djv_image::Image()),
  _image_pick(copy ? copy->_image_pick : true),
  _playback_frame_tmp(0),
  _p(new _Widget),
  _valid(false)
{
  using namespace djv_ui;
  
  DJV_DBG_OBJ("Widget::Widget");
  
  // Widgets.
  
  _p->menu = new menu::Widget;
  
  _p->toolbar = new group::Frame;
  _p->toolbar->edges(false, false, true, true);
  
  _p->window_new = new button::Tool("window_new");
  _p->window_copy = new button::Tool("window_copy");
  _p->window_close = new button::Tool("window_close");
  _p->window_fit = new button::Tool("window_fit");
  _p->window_fullscreen = new button::Tool("window_fullscreen");
  _p->window_fullscreen->type(button::Tool::TOGGLE);
  
  _p->playbar = new group::Frame;
  _p->playbar->edges(false, false, true, false);
  
  _p->view = new view::Widget;

  // Layout.
  
  Layout_V * layout = new Layout_V(this);
  layout->spacing(0);
  layout->margin(0);
  layout->add(_p->menu);
  
  layout->add(_p->toolbar);
  Layout_H * toolbar_layout = new Layout_H(_p->toolbar);
  toolbar_layout->spacing(0);
  toolbar_layout->margin(0);
  
  Layout_H * toolbar_file_layout = new Layout_H(toolbar_layout);
  toolbar_file_layout->spacing(0);
  toolbar_file_layout->margin(0);
  
  toolbar_layout->add(new group::Divider_Horizontal(style::global()->margin()));
  
  Layout_H * toolbar_window_layout = new Layout_H(toolbar_layout);
  toolbar_window_layout->spacing(0);
  toolbar_window_layout->margin(0);
  toolbar_window_layout->add(_p->window_new);
  toolbar_window_layout->add(_p->window_copy);
  toolbar_window_layout->add(_p->window_close);
  toolbar_window_layout->add(_p->window_fit);
  toolbar_window_layout->add(_p->window_fullscreen);
  
  toolbar_layout->add(new group::Divider_Horizontal(style::global()->margin()));
  
  Layout_H * toolbar_view_layout = new Layout_H(toolbar_layout);
  toolbar_view_layout->spacing(0);
  toolbar_view_layout->margin(0);
  
  toolbar_layout->add(new group::Divider_Horizontal(style::global()->margin()));
  
  Layout_H * toolbar_image_layout = new Layout_H(toolbar_layout);
  toolbar_image_layout->spacing(0);
  toolbar_image_layout->margin(0);
  
  toolbar_layout->add_stretch();
  
  layout->add(_p->view);
  layout->stretch(_p->view);
  
  layout->add(_p->playbar);

  // Initialize.
  
  _file = std::auto_ptr<file::Group>(new file::Group(
    _p->menu,
    toolbar_file_layout,
    copy ? copy->_file.get() : 0));
  _view = std::auto_ptr<view::Group>(new view::Group(
    _p->menu,
    toolbar_view_layout,
    copy ? copy->_view.get() : 0));
  _image = std::auto_ptr<image::Group>(new image::Group(
    _p->menu,
    toolbar_image_layout,
    copy ? copy->_image.get() : 0));
  _playback = std::auto_ptr<playback::Group>(new playback::Group(
    _p->menu,
    _p->playbar,
    copy ? copy->_playback.get() : 0));

  if (copy)
  {
    _p->view->view(copy->_p->view->view());
    _p->view->zoom(copy->_p->view->zoom());
  }
  _p->view->hud(_view->hud());
  _p->view->grid(_view->grid());
  
  file_update();
  window_controls_update();
  view_update();
  view_overlay_update();
  image_update();
  group_update();
  
  if (copy)
  {
    size(copy->geom().size);
  }
  else
  {
    if (prefs()->resize_fit())
      window_fit();
    else
      size(prefs()->geom().size);
  }

  // Callbacks.
  
  _file->open_sig.set(this, file_open_callback);
  _file->reload_sig.set(this, file_reload_callback);
  _file->save_sig.set(this, file_save_callback);
  _file->save_frame_sig.set(this, file_save_frame_callback);
  _file->image_sig.set(this, image_callback);
  _file->update_sig.set(this, group_callback);
  
  _p->window_new->sig.set(this, window_new_callback);
  _p->window_copy->sig.set(this, window_copy_callback);
  _p->window_close->sig.set(this, window_close_callback);
  _p->window_fit->sig.set(this, window_fit_callback);
  _p->window_fullscreen->sig.set(this, window_fullscreen_callback);
  
  _view->move_sig.set(this, view_move_callback);
  _view->center_sig.set(this, view_center_callback);
  _view->zoom_sig.set(this, view_zoom_callback);
  _view->zoom_reset_sig.set(this, view_zoom_reset_callback);
  _view->reset_sig.set(this, view_reset_callback);
  _view->fit_sig.set(this, view_fit_callback);
  _view->overlay_sig.set(this, view_overlay_callback);
  _view->update_sig.set(this, group_callback);
  
  _p->view->sig.set(this, view_callback);
  _p->view->pick_sig.set(this, view_pick_callback);
  _p->view->scroll_sig.set(this, playback_frame_shuttle_callback);
  _p->view->scroll_value_sig.set(this, playback_frame_shuttle_value_callback);
  _p->view->file_sig.set(this, file_open_callback);
  
  _image->redraw_sig.set(this, image_redraw_callback);
  _image->resize_sig.set(this, image_resize_callback);
  _image->frame_store_sig.set(this, image_frame_store_callback);
  _image->display_sig.set(this, image_display_callback);
  _image->display_value_sig.set(this, image_display_value_callback);
  _image->magnify_sig.set(this, image_magnify_callback);
  _image->sample_sig.set(this, image_sample_callback);
  _image->histogram_sig.set(this, image_histogram_callback);
  _image->info_sig.set(this, image_info_callback);
  _image->update_sig.set(this, group_callback);

  _playback->image_sig.set(this, image_callback);
  _playback->overlay_sig.set(this, view_overlay_callback);
  _playback->update_sig.set(this, group_callback);
  
  prefs()->toolbar_sig.set(this, window_toolbar_callback);
  prefs()->playbar_sig.set(this, window_playbar_callback);

  close_sig.set(this, window_close2_callback);
  
  _valid = true;
}

Widget::~Widget()
{
  DJV_DBG_OBJ("Widget::~Widget");
}

void Widget::del()
{
  DJV_DBG_OBJ("Widget::del");
  
  prefs()->geom(geom());
  
  djv_ui::window::Widget::del();
  
  image_display::dialog()->del(_p->view);
  image_histogram::dialog()->del(_p->view);
  image_info::dialog()->del(_p->view);
  image_magnify::dialog()->del(_p->view);
  image_sample::dialog()->del(_p->view);
  
  _p->view->set(0);
  
  _file->del();
  _view->del();
  _image->del();
  _playback->del();
}

int Widget::handle(int in)
{
  switch (in)
  {
    case FL_SHORTCUT:
      switch (Fl::event_key())
      {
        case FL_Escape:
          if (fullscreen())
          {
            window_fullscreen(false);
            return 1;
          }
          break;
        
        default:
        
          // XXX
      
          if (! _p->menu->visible())
            if (_p->menu->test_shortcut())
              return 1;
          break;
      }
      break;
  }
  return djv_ui::window::Widget::handle(in);
}

void Widget::dirty()
{
  DJV_DBG_OBJ("Widget::dirty");

  djv_ui::window::Widget::dirty();
  
  if (_valid)
  {
    view_overlay_update();
    image_redraw_update();
    image_dialog_update();
    group_update();
  }
}

void Widget::geom(const Box2i & in)
{
  djv_ui::window::Widget::geom(in);
  
  view_resize_update();
  view_update();
}

const Box2i & Widget::geom() const { return djv_ui::window::Widget::geom(); }

void Widget::window_menu_update(djv_ui::menu::Group * in)
{
  using namespace djv_ui::menu;
  
  const List<djv_ui::shortcut::Shortcut> & shortcut =
    shortcut::prefs()->shortcut();
  
  in->add(menu_title, 0, 0, 0, SUB_MENU);
  
  in->add(
    menu_new, shortcut[shortcut::WINDOW_NEW].value,
    _window_new_callback, this);
  
  in->add(
    menu_copy, shortcut[shortcut::WINDOW_COPY].value,
    _window_copy_callback, this);
  
  in->add(
    menu_close, shortcut[shortcut::WINDOW_CLOSE].value,
    _window_close_callback, this,
    DIVIDER);

  in->add(
    menu_fit, shortcut[shortcut::WINDOW_FIT].value,
    _window_fit_callback, this);
  
  in->add(
    menu_fullscreen, shortcut[shortcut::WINDOW_FULLSCREEN].value,
    _window_fullscreen_callback, this,
    TOGGLE | DIVIDER, _window_fullscreen);
  
  in->add(
    menu_toolbar, shortcut[shortcut::WINDOW_TOOLBAR].value,
    _window_toolbar_callback, this,
    TOGGLE, _window_toolbar);

  in->add(
    menu_playbar, shortcut[shortcut::WINDOW_PLAYBAR].value,
    _window_playbar_callback, this,
    TOGGLE, _window_playbar);

  in->end();
}

void Widget::window_toolbar_update()
{
  using djv_ui::shortcut::serialize;
  
  const List<djv_ui::shortcut::Shortcut> & shortcut =
    shortcut::prefs()->shortcut();
  
  _p->window_new->tooltip(str::Format(tooltip_new).
    arg(serialize(shortcut[shortcut::WINDOW_NEW].value)));
  _p->window_copy->tooltip(str::Format(tooltip_copy).
    arg(serialize(shortcut[shortcut::WINDOW_COPY].value)));
  _p->window_close->tooltip(str::Format(tooltip_close).
    arg(serialize(shortcut[shortcut::WINDOW_CLOSE].value)));
  _p->window_fit->tooltip(str::Format(tooltip_fit).
    arg(serialize(shortcut[shortcut::WINDOW_FIT].value)));
  _p->window_fullscreen->tooltip(str::Format(tooltip_fullscreen).
    arg(serialize(shortcut[shortcut::WINDOW_FULLSCREEN].value)));
}

void Widget::file_open(const File & in)
{
  DJV_DBG_OBJ("Widget::file_open");
  DJV_DBG_PRN("in = " << in);
  
  _image_p = 0;
  _p->view->set(0);

  callbacks(false);
  
  _file->open(in);
  
  const djv_image::Time & time = _file->time();
  _playback->time(time);
  _playback->frame(0);
  _playback->playback(
    (time.list.size() > 1 && playback::prefs()->start()) ?
    playback::FORWARD : playback::STOP);
  
  callbacks(true);
  
  file_update();
  image_update();

  if (prefs()->resize_fit())
  {
    if (! _window_fullscreen)
    {
      _p->view->view_zero();
      window_fit();
    }
    else
    {
      _p->view->view_fit();
    }
  }
  else
  {
    view_resize_update();
  }
  view_update();

  group_update();
}

void Widget::file_open_callback(const File & in) { file_open(in); }

void Widget::file_reload_callback(bool)
{
  if (app()->image_prefs()->frame_store_file_reload())
    image_frame_store_callback(true);
}

void Widget::file_save_callback(const File & in)
{
  djv_image::Time time;
  time.list = List<int>(
    _playback->time().list,
    _playback->in(),
    _playback->out() - _playback->in() + 1);
  time.speed = _playback->speed();
  
  file::save()->set(file::Save_Info(
    _file->file(),
    in,
    _file->info()[_file->layer()],
    _file->layer(),
    _file->proxy(),
    time,
    image_option()));
}

void Widget::file_save_frame_callback(const File & in)
{
  djv_image::Time time;
  time.list += _playback->time().list[_playback->frame()];
  time.speed = _playback->speed();
  
  file::save()->set(file::Save_Info(
    _file->file(),
    in,
    _file->info()[_file->layer()],
    _file->layer(),
    _file->proxy(),
    time,
    image_option()));
}

void Widget::file_update()
{
  DJV_DBG_OBJ("Widget::file_update");

  callbacks(false);
  
  const File & file = _file->file();
  DJV_DBG_PRN("file = " << file);
  
  title(! file.get().empty() ?
    str::Format("%% - %%").arg(app()->name()).arg(file) :
    str::Format("%%").arg(app()->name()));

  callbacks(true);
}

void Widget::file_layer(int in) { _file->layer(in); }
void Widget::file_proxy(djv_image::data::PROXY in) { _file->proxy(in); }
void Widget::file_cache(bool in) { _file->cache(in); }

void Widget::window_new_callback(bool)
{
  DJV_DBG_OBJ("Widget::window_new_callback");

  factory()->add()->show();
}

void Widget::_window_new_callback() { window_new_callback(true); }

void Widget::window_copy_callback(bool)
{
  DJV_DBG_OBJ("Widget::window_copy_callback");

  factory()->add(this)->show();
}

void Widget::_window_copy_callback() { window_copy_callback(true); }

void Widget::window_close_callback(bool)
{
  if (1 == factory()->list().size()) return;
  
  DJV_DBG_OBJ("Widget::window_close_callback");

  window_close2_callback(true);
}

void Widget::window_close2_callback(bool) { factory()->del(this); }

void Widget::_window_close_callback() { window_close_callback(true); }

void Widget::window_fit()
{
  DJV_DBG_OBJ("Widget::window_fit");
  
  // Image size.
  
  const Box2f bbox = _p->view->bbox().size;
  DJV_DBG_PRN("bbox = " << bbox);
    
  Vec2i image_size = vec::ceil<double, int>(bbox.size);
  if (! vec::size_valid(image_size))
  {
    image_size = Vec2i(600, 300);
  }
  DJV_DBG_PRN("image size = " << image_size);

  // Fit to screen size.
  
  const double resize_max = window::resize_max(
    prefs()->resize_fit() ? prefs()->resize_max() : RESIZE_MAX_NONE);
  
  const Vec2i max(int(Fl::w() * resize_max), int(Fl::h() * resize_max));
  if (image_size.x > max.x || image_size.y > max.y)
    image_size = (image_size.x > image_size.y) ?
      Vec2i(max.x, math::ceil(image_size.y / double(image_size.x) * max.x)) :
      Vec2i(math::ceil(image_size.x / double(image_size.y) * max.y), max.y);

  // Size.

  const Vec2i ui_size = size_hint() - Vec2i(0, _p->view->size_hint().y);
  DJV_DBG_PRN("ui size = " << ui_size);
  
  const Vec2i size = Vec2i(
    math::max(image_size.x, ui_size.x),
    image_size.y + ui_size.y);
  DJV_DBG_PRN("size = " << size);
  
  fullscreen(false);
  _window_fullscreen = false;
  this->size(size);
  _p->view->view_fit();
  _p->view->redraw();
  view_update();
}

void Widget::window_fit_callback(bool) { window_fit(); }

void Widget::_window_fit_callback() { window_fit_callback(true); }

void Widget::window_fullscreen(bool in)
{
  if (in == _window_fullscreen) return;
  
  DJV_DBG_OBJ("Widget::window_fullscreen");
  DJV_DBG_PRN("in = " << in);
  
  _window_fullscreen = in;
  fullscreen(_window_fullscreen);
  
  window_fullscreen_update();
  group_update();
  update();
  if (prefs()->resize_fit()) _p->view->view_fit();
  else view_resize_update();
  view_update();
}

void Widget::window_fullscreen_callback(bool in) { window_fullscreen(in); }

void Widget::_window_fullscreen_callback()
{
  window_fullscreen(_p->menu->value());
}

void Widget::window_fullscreen_update()
{
  DJV_DBG_OBJ("Widget::window_fullscreen_update");

  callbacks(false);
  
  _p->window_fullscreen->set(_window_fullscreen);
  
  if (_window_fullscreen)
    _p->menu->hide();
  else
    _p->menu->show();
  
  if (_window_fullscreen)
  {
    _p->toolbar->hide();
    _p->playbar->hide();
  }
  else
  {
    if (_window_toolbar) _p->toolbar->show();
    if (_window_playbar) _p->playbar->show();
  }
  
  callbacks(true);
}

void Widget::window_toolbar(bool in)
{
  if (in == _window_toolbar) return;
  
  DJV_DBG_OBJ("Widget::window_toolbar");
  DJV_DBG_PRN("in = " << in);
  
  _window_toolbar = in;
  
  window_controls_update();
  //window_resize_update();
  group_update();
  update();
  window_resize_update(); // XXX Needs update() first.
}

void Widget::window_toolbar_callback(bool in) { window_toolbar(in); }
void Widget::_window_toolbar_callback() { window_toolbar(_p->menu->value()); }

void Widget::window_playbar(bool in)
{
  if (in == _window_playbar) return;
  
  DJV_DBG_OBJ("Widget::window_playbar");
  DJV_DBG_PRN("in = " << in);
  
  _window_playbar = in;
  
  window_controls_update();
  //window_resize_update();
  group_update();
  update();
  window_resize_update(); // XXX Needs update() first.
}

void Widget::window_playbar_callback(bool in) { window_playbar(in); }
void Widget::_window_playbar_callback() { window_playbar(_p->menu->value()); }

void Widget::window_controls_update()
{
  DJV_DBG_OBJ("Widget::window_controls_update");

  callbacks(false);
  
  if (_window_toolbar && ! _window_fullscreen)
    _p->toolbar->show();
  else
    _p->toolbar->hide();

  if (_window_playbar && ! _window_fullscreen)
    _p->playbar->show();
  else
    _p->playbar->hide();

  callbacks(true);
}

void Widget::window_resize_update()
{
  DJV_DBG_OBJ("Widget::window_resize_update");
  DJV_DBG_PRN("geom = " << geom());
  
  callbacks(false);
  
  if (prefs()->resize_fit())
  {
    if (! _window_fullscreen)
    {
      window_fit();
    }
    else
    {
      _p->view->view_fit();
    }
  }
  
  callbacks(true);
}

void Widget::view_grid(view::GRID in) { _view->grid(in); }
void Widget::view_hud(bool in) { _view->hud(in); }

void Widget::view_callback(bool)
{
  DJV_DBG_OBJ("Widget::view_callback");
  
  view_update();
}

void Widget::view_move_callback(const Vec2i & in)
{
  _p->view->view(_p->view->view() + in);
  _p->view->redraw();
}

void Widget::view_center_callback(bool)
{
  _p->view->view_center();
  _p->view->redraw();
}

void Widget::view_zoom_callback(double in)
{
  _p->view->zoom_focus(_p->view->zoom() * in);
  _p->view->redraw();
}

void Widget::view_zoom_reset_callback(bool)
{
  _p->view->zoom_focus(1.0);
  _p->view->redraw();
}

void Widget::view_reset_callback(bool)
{
  _p->view->view_zero();
  _p->view->redraw();
}

void Widget::view_fit_callback(bool)
{
  _p->view->view_fit();
  _p->view->redraw();
}

void Widget::view_update()
{
  DJV_DBG_OBJ("Widget::view_update");

  callbacks(false);

  _view->zoom(_p->view->zoom());

  _image->pick(_image_pick, _p->view->zoom());

  callbacks(true);
}

void Widget::view_resize_update()
{
  DJV_DBG_OBJ("Widget::view_resize_update");

  callbacks(false);
  
  switch (view::prefs()->resize())
  {
    case view::RESIZE_NONE: break;
    case view::RESIZE_FIT: _p->view->view_fit(); break;
    case view::RESIZE_CENTER: _p->view->view_center(); break;
  }

  callbacks(true);
}

void Widget::view_overlay_callback(bool) { view_overlay_update(); }

void Widget::view_overlay_update()
{
  DJV_DBG_OBJ("Widget::view_overlay_update");
  
  callbacks(false);

  _p->view->grid(_view->grid());
  _p->view->grid_color(view::prefs()->grid_color());

  view::Hud hud;
  if (_image_p)
  {
    hud.info = _image_p->info();
    hud.tag = _image_p->tag;
  }
  const djv_image::Time & time = _playback->time();
  hud.frame = time.list.size() ? time.list[_playback->frame()] : 0;
  hud.speed = _playback->speed();
  hud.speed_real = _playback->speed_real();
  hud.dropped_frames = _playback->dropped_frames();
  _p->view->hud(_view->hud());
  _p->view->hud_info(hud);
  _p->view->hud_option(view::prefs()->hud_option());
  _p->view->hud_background(view::prefs()->hud_background());
  _p->view->hud_color_foreground(view::prefs()->hud_color_foreground());
  _p->view->hud_color_background(view::prefs()->hud_color_background());
  
  _p->view->redraw();

  // XXX

  _playback->cache(
    file::prefs()->cache_display() ? _file->cache_list() : List<int>());

  callbacks(true);
}

void Widget::view_pick_callback(const Vec2i & in)
{
  DJV_DBG_OBJ("Widget::view_pick_callback");
  DJV_DBG_PRN("in = " << in);
  
  _image_pick = in - _p->view->view();
  
  image_pick_update();
}
  
djv_image::gl_data::Option Widget::image_option() const
{
  djv_image::gl_data::Option out;

  out.xform.mirror = _image->mirror();
  if (_image_p)
    out.xform.scale = image::scale(_image->scale(), _image_p->size());
  out.xform.rotate = image::rotate(_image->rotate());

  if (_image->profile() && _image_p)
    out.profile = _image_p->profile;
  out.channel = _image->channel();
  out.display = _image->display();
  out.background = view::prefs()->background();
  
  return out;
}

void Widget::image_callback(bool)
{
  image_update();
}

void Widget::image_update()
{
  DJV_DBG_OBJ("Widget::image_update");
  
  callbacks(false);

  _image_p = _file->get(_playback->frame());
  if (_image_p) DJV_DBG_PRN("image = " << *_image_p);

  callbacks(true);
  
  view_overlay_update();
  image_redraw_update();
}

void Widget::image_redraw_callback(bool)
{
  image_redraw_update();
}

void Widget::image_redraw_update()
{
  DJV_DBG_OBJ("Widget::image_redraw_update");
  DJV_DBG_PRN("view = " << _p->view->view());
  DJV_DBG_PRN("zoom = " << _p->view->zoom());
  
  callbacks(false);

  const djv_image::gl_data::Option & option = image_option();
  DJV_DBG_PRN("color = " << option.display.color);
  
  _p->view->set(_image->frame_store() ? &_image_tmp : _image_p);
  _p->view->option(option);
  _p->view->redraw();

  callbacks(true);
  
  image_dialog_update();
}

void Widget::image_resize_callback(bool) { window_resize_update(); }

void Widget::image_frame_store_callback(bool)
{
  DJV_DBG_OBJ("Widget::image_frame_store_callback");
  
  if (_image_p) _image_tmp = *_image_p;
}

void Widget::image_display_callback(bool)
{
  image_display::dialog()->show();
  image_display::dialog()->set(_p->view, _image->display());
  image_display::dialog()->sig.del();
  image_display::dialog()->sig.set(this, image_display_value_callback);
}

void Widget::image_display_value_callback(const image::Display & in)
{
  DJV_DBG_OBJ("Widget::image_display_value_callback");
  
  _image->display(in);

  image_redraw_update();
  image_dialog_update();
}

void Widget::image_histogram_callback(bool)
{
  image_histogram::dialog()->show();
  image_histogram::dialog()->set(_p->view);
}

void Widget::image_info_callback(bool)
{
  image_info::dialog()->show();
  image_info::dialog()->set(_p->view, image_info::Info(_image_p));
}

void Widget::image_magnify_callback(bool)
{
  image_magnify::dialog()->show();
  image_magnify::dialog()->set(_p->view, _image_pick);
}

void Widget::image_sample_callback(bool)
{
  image_sample::dialog()->show();
  image_sample::dialog()->set(_p->view, _image_pick);
}

void Widget::image_dialog_update()
{
  DJV_DBG_OBJ("Widget::image_dialog_update");
  
  callbacks(false);

  image_display::dialog()->update(_p->view, _image->display());
  image_histogram::dialog()->update(_p->view);
  image_info::dialog()->update(_p->view, image_info::Info(_image_p));
  image_magnify::dialog()->update(_p->view);
  image_sample::dialog()->update(_p->view);

  callbacks(true);
}

void Widget::image_pick_update()
{
  DJV_DBG_OBJ("Widget::image_pick_update");
  
  callbacks(false);
  
  _image->pick(_image_pick, _p->view->zoom());
  
  image_display::dialog()->pick(_p->view, _image->display());
  image_display::dialog()->sig.del();
  image_display::dialog()->sig.set(this, image_display_value_callback);
  image_histogram::dialog()->pick(_p->view);
  image_info::dialog()->pick(_p->view, image_info::Info(_image_p));
  image_magnify::dialog()->pick(_p->view, _image_pick);
  image_sample::dialog()->pick(_p->view, _image_pick);
  
  callbacks(true);
}

void Widget::playback(playback::PLAYBACK in) { _playback->playback(in); }
void Widget::playback_frame(int in) { _playback->frame(in); }
void Widget::playback_speed(const djv_image::time::Speed & in)
  { _playback->speed(in); }

void Widget::playback_frame_shuttle_callback(bool)
{
  _playback->playback(playback::STOP);
  _playback_frame_tmp = _playback->frame();
}

void Widget::playback_frame_shuttle_value_callback(int in)
{
  _playback->frame(_playback_frame_tmp + in);
}
  
void Widget::group_callback(bool) { group_update(); }

void Widget::group_update()
{
  DJV_DBG_OBJ("Widget::group_update");

  callbacks(false);
  
  _file->toolbar_update();
  window_toolbar_update();
  _view->toolbar_update();
  _image->toolbar_update();
  _playback->playbar_update();
  
  djv_ui::menu::Group group;
  _file->menu_update(&group);
  window_menu_update(&group);
  _view->menu_update(&group);
  _image->menu_update(&group);
  _playback->menu_update(&group);
  _p->menu->set(group);
  
  callbacks(true);
}

}
}

