/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_program.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "gl_data_program.h"

namespace djv_image {
namespace gl_data {
namespace program {

namespace {

const Str
  error_name = "OpenGL",
  error_texture_max = "Maximum Texture Size: %%";

const Str src_header =
"#version 110\n"
"#extension GL_ARB_texture_rectangle : enable\n"
"\n"
"struct Levels { float in0, in1, gamma, out0, out1; };\n"
"struct Exposure { float v, d, k, f, g; };\n"
"\n"
"uniform sampler2DRect texture;\n"
"\n"
"uniform struct\n"
"{\n"
"  sampler1D lut;\n"
"  Exposure exposure;\n"
"} profile;\n"
"\n"
"uniform struct\n"
"{\n"
"  sampler1D lut;\n"
"  mat4 color;\n"
"  Levels levels;\n"
"  float soft_clip;\n"
"} display;\n"
"\n"
"uniform sampler2DRect scale_contrib;\n"
"\n"
"float knee(float value, float f)\n"
"{\n"
"  return log(value * f + 1.0) / f;\n"
"}\n"
"\n";

const Str src_lut =
"vec4 lut(vec4 value, sampler1D lut)\n"
"{\n"
"  value[0] = texture1D(lut, value[0])[0];\n"
"  value[1] = texture1D(lut, value[1])[1];\n"
"  value[2] = texture1D(lut, value[2])[2];\n"
"  return value;\n"
"}\n"
"\n";

const Str src_display_color =
"vec4 display_color(vec4 value, mat4 color)\n"
"{\n"
"  vec4 tmp;\n"
"  tmp[0] = value[0];\n"
"  tmp[1] = value[1];\n"
"  tmp[2] = value[2];\n"
"  tmp[3] = 1.0;\n"
"  tmp *= color;\n"
"  tmp[3] = value[3];\n"
"  return tmp;\n"
"}\n"
"\n";

const Str src_levels =
"vec4 levels(vec4 value, Levels data)\n"
"{\n"
"  value[0] = pow(max(value[0] - data.in0, 0.0) / data.in1, data.gamma) *\n"
"    data.out1 + data.out0;\n"
"  value[1] = pow(max(value[1] - data.in0, 0.0) / data.in1, data.gamma) *\n"
"    data.out1 + data.out0;\n"
"  value[2] = pow(max(value[2] - data.in0, 0.0) / data.in1, data.gamma) *\n"
"   data.out1 + data.out0;\n"
"  return value;\n"
"}\n"
"\n";

const Str src_exposure =
"vec4 exposure(vec4 value, Exposure data)\n"
"{\n"
"  value[0] = max(0.0, value[0] - data.d) * data.v;\n"
"  value[1] = max(0.0, value[1] - data.d) * data.v;\n"
"  value[2] = max(0.0, value[2] - data.d) * data.v;\n"
"\n"
"  if (value[0] > data.k)\n"
"    value[0] = data.k + knee(value[0] - data.k, data.f);\n"
"  if (value[1] > data.k)\n"
"    value[1] = data.k + knee(value[1] - data.k, data.f);\n"
"  if (value[2] > data.k)\n"
"    value[2] = data.k + knee(value[2] - data.k, data.f);\n"
"\n"
"  value[0] = pow(value[0], data.g) * 0.332;\n"
"  value[1] = pow(value[1], data.g) * 0.332;\n"
"  value[2] = pow(value[2], data.g) * 0.332;\n"
"  return value;\n"
"}\n"
"\n";

const Str src_soft_clip =
"vec4 soft_clip(vec4 value, float soft_clip)\n"
"{\n"
"  float tmp = 1.0 - soft_clip;\n"
"  if (value[0] > tmp)\n"
"    value[0] = tmp + (1.0 - exp(-(value[0] - tmp) / soft_clip)) * soft_clip;\n"
"  if (value[1] > tmp)\n"
"    value[1] = tmp + (1.0 - exp(-(value[1] - tmp) / soft_clip)) * soft_clip;\n"
"  if (value[2] > tmp)\n"
"    value[2] = tmp + (1.0 - exp(-(value[2] - tmp) / soft_clip)) * soft_clip;\n"
"  return value;\n"
"}\n"
"\n";

const Str src_scale_x =
"int scale_size = %%;\n"
"\n"
"vec4 scale_x(sampler2DRect texture, sampler2DRect contrib)\n"
"{\n"
"  vec4 value = vec4(0.0);\n"
"  for (int i = 0; i < scale_size; ++i)\n"
"  {\n"
"    vec4 tmp = texture2DRect(\n"
"      contrib, vec2(gl_TexCoord[0].s, float(i)));\n"
"    vec2 position = vec2(tmp[0] + 0.5, gl_TexCoord[0].t);\n"
"    value += tmp[3] * %%;\n"
"  }\n"
"  return value;\n"
"}\n"
"\n";

const Str src_scale_y =
"int scale_size = %%;\n"
"\n"
"vec4 scale_y(sampler2DRect texture, sampler2DRect contrib)\n"
"{\n"
"  vec4 value = vec4(0.0);\n"
"  for (int i = 0; i < scale_size; ++i)\n"
"  {\n"
"    vec4 tmp = texture2DRect(\n"
"      contrib, vec2(gl_TexCoord[0].t, float(i)));\n"
"    vec2 position = vec2(gl_TexCoord[0].s, tmp[0] + 0.5);\n"
"    value += tmp[3] * texture2DRect(texture, position);\n"
"  }\n"
"  return value;\n"
"}\n"
"\n";

const Str src_sample =
"texture2DRect(texture, position)\n";
const Str src_sample_lut =
"lut(texture2DRect(texture, position), profile.lut)\n";
const Str src_sample_exposure =
"exposure(texture2DRect(texture, position), profile.exposure)\n";

const Str src_filter_default =
"vec2 position = gl_TexCoord[0].st;\n"
"color = %%;\n";
const Str src_filter_x =
"color = scale_x(texture, scale_contrib);\n";
const Str src_filter_y =
"color = scale_y(texture, scale_contrib);\n";

const Str src_main =
"void main(void)\n"
"{\n"
"  vec4 color;\n"
"%%"
"  gl_FragColor = color;\n"
"}\n";

Str src(profile::PROFILE profile, CHANNEL channel, const Display & display,
  Filter::FILTER filter, int scale_size, bool scale_x, bool clamp)
{
  DJV_DBG_OBJ("src");
  DJV_DBG_PRN("clamp = " << clamp);
  
  Str header, main;
  
  header = src_header;
  header += src_lut;

  Str sample;
  switch (profile)
  {
    case profile::LUT:
      sample = src_sample_lut;
      break;
    case profile::EXPOSURE:
      header += src_exposure;
      sample = src_sample_exposure;
      break;
    default:
      sample = src_sample;
      break;
  }
  
  switch (filter)
  {
    case Filter::FILTER_DEFAULT:
      main += str::Format(src_filter_default).arg(sample);
      break;
    case Filter::FILTER_CUSTOM:
      header += str::Format(scale_x ? src_scale_x : src_scale_y).
        arg(scale_size).
        arg(sample);
      main += scale_x ? src_filter_x : src_filter_y;
      break;
  }
  
  if (channel)
    main += str::Format("color = vec4(color[%%]);\n").arg(channel - 1);
  
  if (vec::size_valid(display.lut.size()))
    main += "color = lut(color, display.lut);\n";
  if (display.color != Display().color)
  {
    header += src_display_color;
    main += "color = display_color(color, display.color);\n";
  }
  if (display.levels != Display().levels)
  {
    header += src_levels;
    main += "color = levels(color, display.levels);\n";
  }
  if (display.soft_clip != Display().soft_clip)
  {
    header += src_soft_clip;
    main += "color = soft_clip(color, display.soft_clip);\n";
  }
  
  if (clamp)
    main += "color = clamp(color, vec4(0.0), vec4(1.0));\n";
  
  return header + "\n" + Str(str::Format(src_main).arg(main));
}

double knee(double x, double f)
{
  return math::log(x * f + 1.0) / f;
}

double knee_f(double x, double y)
{
  double f0 = 0.0, f1 = 1.0;

  while (knee(x, f1) > y)
  {
    f0 = f1;
    f1 = f1 * 2.0;
  }

  for (int i = 0; i < 30; ++i)
  {
    const double f2 = (f0 + f1) / 2.0;
    if (knee(x, f2) < y) f1 = f2;
    else f0 = f2;
  }

  return (f0 + f1) / 2.0;
}

void var_profile(const Option & option, GLhandleARB program, Lut * profile)
{
  DJV_DBG_OBJ("var_profile");

  if (profile::LUT == option.profile.type)
  {
    DJV_DBG_GL(glActiveTexture(GL_TEXTURE2));
    glUniform1i(glGetUniformLocation(program, "profile.lut"), 2);
    profile->init(option.profile.lut);
  }
  
  struct Exposure { double v, d, k, f, g; } exposure;
  exposure.v = math::pow(2.0, option.profile.exposure.value + 2.47393);
  exposure.d = option.profile.exposure.defog;
  exposure.k = math::pow(2.0, option.profile.exposure.knee_low);
  exposure.f = knee_f(math::pow(2.0, option.profile.exposure.knee_high) -
    exposure.k, math::pow(2.0, 3.5) - exposure.k);
  exposure.g = 1.0 / option.profile.exposure.gamma;
  DJV_DBG_PRN("exposure v = " << exposure.v);
  DJV_DBG_PRN("exposure d = " << exposure.d);
  DJV_DBG_PRN("exposure k = " << exposure.k);
  DJV_DBG_PRN("exposure f = " << exposure.f);
  DJV_DBG_PRN("exposure g = " << exposure.g);
  glUniform1f(
    glGetUniformLocation(program, "profile.exposure.v"),
    exposure.v);
  glUniform1f(
    glGetUniformLocation(program, "profile.exposure.d"),
    exposure.d);
  glUniform1f(
    glGetUniformLocation(program, "profile.exposure.k"),
    exposure.k);
  glUniform1f(
    glGetUniformLocation(program, "profile.exposure.f"),
    exposure.f);
  glUniform1f(
    glGetUniformLocation(program, "profile.exposure.g"),
    exposure.g);
}

void var_display(const Option & option, GLhandleARB program, Lut * display)
{
  DJV_DBG_OBJ("var_display");

  glUniformMatrix4fv(
    glGetUniformLocation(program, "display.color"), 1, false,
    mtx::convert<double, GLfloat>(
      mtx::transpose(color_mtx(option.display.color))).e);

  glUniform1f(
    glGetUniformLocation(program, "display.levels.in0"),
    option.display.levels.in_low);
  glUniform1f(
    glGetUniformLocation(program, "display.levels.in1"),
    option.display.levels.in_high - option.display.levels.in_low);
  glUniform1f(
    glGetUniformLocation(program, "display.levels.gamma"),
    1.0 / option.display.levels.gamma);
  glUniform1f(
    glGetUniformLocation(program, "display.levels.out0"),
    option.display.levels.out_low);
  glUniform1f(
    glGetUniformLocation(program, "display.levels.out1"),
    option.display.levels.out_high - option.display.levels.out_low);
  
  glUniform1f(
    glGetUniformLocation(program, "display.soft_clip"),
    option.display.soft_clip);
  
  if (vec::size_valid(option.display.lut.size()))
  {
    DJV_DBG_GL(glActiveTexture(GL_TEXTURE3));
    glUniform1i(glGetUniformLocation(program, "display.lut"), 3);
    display->init(option.display.lut);
  }
}

}

void Program::draw(const djv_image::Data & in, const Option & option)
  throw (Error)
{
  if (! GLEW_VERSION_1_3 || ! GLEW_ARB_shader_objects) return;
  
  DJV_DBG_OBJ("Program::draw");
  DJV_DBG_PRN("in = " << in);

  const Info & info = in.info();
  const int proxy_scale = djv_image::data::proxy_scale(info.proxy);
  const Vec2i scale = vec::ceil<double, int>(
    option.xform.scale * Vec2f(info.size * proxy_scale));
  const Vec2i scale_tmp(scale.x, in.h());
  DJV_DBG_PRN("scale = " << scale);

  Filter::FILTER filter = option.filter.value;
  if (info.size == scale)
  {
    // XXX Don't use multi-pass for identity scale.
    
    filter = Filter::FILTER_DEFAULT;
  }

  // Initialize.
  
  if (Filter::FILTER_DEFAULT == filter)
  {
    _texture.init(in, filter_to_gl(option.filter._default.first),
      filter_to_gl(option.filter._default.second));

    const State_Default state(option);
    if (_state_default != state)
    {
      DJV_DBG_PRN("init default");

      _shader.init(src(option.profile.type, option.channel, option.display,
        Filter::FILTER_DEFAULT, 0, false, option.clamp));
      
      _state_default = state;
    }
  }
  else
  {
    _texture.init(in.info(), GL_NEAREST, GL_NEAREST);

    const State_Custom state(scale, proxy_scale, option);
    if (_state_custom != state)
    {
      DJV_DBG_PRN("init custom");

      _buffer = std::auto_ptr<gl_offscreen::Buffer>(
        gl_offscreen::global()->buffer_create());
    
      djv_image::Data scale_contrib;
      program::scale_contrib(in.w(), scale.x, option.filter.custom,
        &scale_contrib);
      _scale_x_contrib.init(scale_contrib, GL_NEAREST, GL_NEAREST);
      _scale_x_shader.init(src(option.profile.type, CHANNEL(0), Display(),
        Filter::FILTER_CUSTOM, scale_contrib.h(), true, false));

      program::scale_contrib(in.h(), scale.y, option.filter.custom,
        &scale_contrib);
      _scale_y_contrib.init(scale_contrib, GL_NEAREST, GL_NEAREST);
      _scale_y_shader.init(src(profile::PROFILE(0), option.channel,
        option.display, Filter::FILTER_CUSTOM, scale_contrib.h(), false,
        option.clamp));
        
      _state_custom = state;
    }
  }
  
  // Render.
  
  const Vec2b mirror(
    info.mirror.x ? (! option.xform.mirror.x) : option.xform.mirror.x,
    info.mirror.y ? (! option.xform.mirror.y) : option.xform.mirror.y);
  DJV_DBG_PRN("mirror = " << mirror);

  if (Filter::FILTER_DEFAULT == filter)
  {
    _shader.bind();
    var_profile(option, _shader.program(), &_lut_profile);
    var_display(option, _shader.program(), &_lut_display);
  
    glActiveTexture(GL_TEXTURE0);
    glUniform1i(glGetUniformLocation(_shader.program(), "texture"), 0);
    _texture.bind();

    DJV_DBG_GL(glPushMatrix());
    DJV_DBG_GL(glLoadMatrixd(mtx::mtx4(xform_mtx(option.xform)).e));  
    quad(in.size(), mirror, proxy_scale);
    DJV_DBG_GL(glPopMatrix());
    
    _shader.unbind();
  }
  else
  {
    // Horizontal pass.
  
    _buffer->init(Info(scale_tmp, in.pixel()));
    _buffer->bind();
  
    _scale_x_shader.bind();
    var_profile(option, _scale_x_shader.program(), &_lut_profile);
    
    glActiveTexture(GL_TEXTURE0);
    glUniform1i(glGetUniformLocation(_scale_x_shader.program(), "texture"), 0);
    _texture.bind();
    _texture.copy(in);
  
    glActiveTexture(GL_TEXTURE1);
    glUniform1i(glGetUniformLocation(_scale_x_shader.program(),
      "scale_contrib"), 1);
    _scale_x_contrib.bind();

    glPushAttrib(GL_TRANSFORM_BIT | GL_VIEWPORT_BIT);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
  
    djv_image::gl::ortho(scale_tmp);
    glViewport(0, 0, scale_tmp.x, scale_tmp.y);
    quad(scale_tmp, mirror);

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    glPopAttrib();

    _scale_x_shader.unbind();
    _buffer->unbind();

    // Vertical pass.
  
    _scale_y_shader.bind();
    var_display(option, _scale_y_shader.program(), &_lut_display);
  
    glActiveTexture(GL_TEXTURE0);
    glUniform1i(glGetUniformLocation(_scale_y_shader.program(), "texture"), 0);
    _buffer->texture_bind();
  
    glActiveTexture(GL_TEXTURE1);
    glUniform1i(glGetUniformLocation(_scale_y_shader.program(),
      "scale_contrib"), 1);
    _scale_y_contrib.bind();

    Xform xform = option.xform;
    xform.scale = Vec2f(1.0);
    const Mtx3f mtx = xform_mtx(xform);
    DJV_DBG_GL(glPushMatrix());
    DJV_DBG_GL(glLoadMatrixd(mtx::mtx4(mtx).e));  
    quad(scale);
    DJV_DBG_GL(glPopMatrix());
    
    _scale_y_shader.unbind();
  }
}

}
}
}
