/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_program_scale.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include "gl_data_program.h"

namespace djv_image {
namespace gl_data {
namespace program {

namespace {

typedef double (Filter_Fnc)(const double t);

static const double support_box = 0.5;

static double filter_box(double t)
{
  if (t > -0.5 && t <= 0.5) return 1.0;
  return 0.0;
}

static const double support_triangle = 1.0;

static double filter_triangle(double t)
{
  if (t < 0.0) t = -t;
  if (t < 1.0) return 1.0 - t;
  return 0.0;
}

static const double support_bell = 1.5;

static double filter_bell(double t)
{
  if (t < 0.0) t = -t;
  if (t < 0.5) return 0.75 - t * t;
  if (t < 1.5)
  {
    t = t - 1.5;
    return 0.5 * t * t;
  }
  return 0.0;
}

static const double support_bspline = 2.0;

static double filter_bspline(double t)
{
  if (t < 0.0) t = -t;
  if (t < 1.0)
  {
    const double tt = t * t;
    return (0.5 * tt * t) - tt + 2.0 / 3.0;
  }
  else if (t < 2.0)
  {
    t = 2.0 - t;
    return (1.0 / 6.0) * (t * t * t);
  }
  return 0.0;
}

static double sinc(double x)
{
  x *= math::pi;
  if (x != 0.0) return math::sin(x) / x;
  return 1.0;
}

static const double support_lanczos3 = 3.0;

static double filter_lanczos3(double t)
{
  if (t < 0.0) t = -t;
  if (t < 3.0) return sinc(t) * sinc(t / 3.0);
  return 0.0;
}

static const double support_cubic = 1.0;

static double filter_cubic(double t)
{
  if (t < 0.0) t = -t;
  if (t < 1.0) return (2.0 * t - 3.0) * t * t + 1.0;
  return 0.0;
}

static const double support_mitchell = 2.0;

static double filter_mitchell(double t)
{
  const double tt = t * t;
  static const double b = 1.0 / 3.0;
  static const double c = 1.0 / 3.0;
  if (t < 0.0) t = -t;
  if (t < 1.0)
  {
    t =
      ((12.0 - 9.0 * b - 6.0 * c) * (t * tt)) +
      ((-18.0 + 12.0 * b + 6.0 * c) * tt) +
      (6.0 - 2.0 * b);
    return t / 6.0;
  }
  else if (t < 2.0)
  {
    t =
      ((-1.0 * b - 6.0 * c) * (t * tt)) +
      ((6.0 * b + 30.0 * c) * tt) +
      ((-12.0 * b - 48.0 * c) * t) +
      (8.0 * b + 24.0 * c);
    return t / 6.0;
  }
  return 0.0;
}

Filter_Fnc * filter_fnc(Filter::CUSTOM in)
{
  static Filter_Fnc * tmp [] =
  {
    filter_box,
    filter_triangle,
    filter_bell,
    filter_bspline,
    filter_lanczos3,
    filter_cubic,
    filter_mitchell
  };
  return tmp[in];
}

static double filter_support(Filter::CUSTOM in)
{
  static const double tmp [] =
  {
    support_box,
    support_triangle,
    support_bell,
    support_bspline,
    support_lanczos3,
    support_cubic,
    support_mitchell
  };
  return tmp[in];
}

int edge(int in, int size)
{
  return math::clamp(in, 0, size - 1);
}

}

void scale_contrib(
  int input,
  int output,
  const Filter::Custom & filter,
  djv_image::Data * out
) {
  DJV_DBG_OBJ("scale_contrib");
  DJV_DBG_PRN("scale = " << input << " " << output);
  DJV_DBG_PRN("filter = " << filter);

  // Filter function.

  const Filter::CUSTOM f = output < input ? filter.first : filter.second;
  Filter_Fnc * fnc = filter_fnc(f);
  const double support = filter_support(f);
  DJV_DBG_PRN("filter = " << f);
  DJV_DBG_PRN("support = " << support);

  // XXX Edges?

  //const double scale = double(output) / double(input);
  const double scale =
    (output > 1 ? (output - 1) : 1) / double(input > 1 ? (input - 1) : 1);
  DJV_DBG_PRN("scale = " << scale);

  const double radius = support * (scale >= 1.0) ? 1.0 : (1.0 / scale);
  DJV_DBG_PRN("radius = " << radius);

  // Initialize.

  const int width = math::ceil(radius * 2 + 1);
  //DJV_DBG_PRN("width = " << width);
  
  out->set(djv_image::Info(Vec2i(output, width), djv_image::pixel::LA_F32));

  // Work.
  
  for (int i = 0; i < output; ++i)
  {
    const double center = i / scale;
    int left = math::ceil(center - radius);
    int right = math::floor(center + radius);
    //DJV_DBG_PRN(i << " = " << left << " " << center << " " << right);
    
    double sum = 0.0;
    int pixel = 0;
    int j = 0;
    for (int k = left; j < width && k <= right; ++j, ++k)
    {
      djv_image::pixel::F32_T * p =
        reinterpret_cast<djv_image::pixel::F32_T *>(out->data(i, j));

      pixel = edge(k, input);
      const double x = (center - k) * (scale < 1.0 ? scale : 1.0);
      const double w = (scale < 1.0) ? ((*fnc)(x) * scale) : (*fnc)(x);
      //DJV_DBG_PRN("w = " << w);
      
      p[0] = pixel;
      p[1] = w;
      
      sum += w;
    }
    for (; j < width; ++j)
    {
      djv_image::pixel::F32_T * p =
        reinterpret_cast<djv_image::pixel::F32_T *>(out->data(i, j));
      
      p[0] = pixel;
      p[1] = 0.0;
    }
    
    /*for (j = 0; j < width; ++j)
    {
      djv_image::pixel::F32_T * p =
        reinterpret_cast<djv_image::pixel::F32_T *>(out->data(i, j));
      DJV_DBG_PRN(p[0] << " = " << p[1]);
    }
    DJV_DBG_PRN("sum = " << sum);*/

    for (j = 0; j < width; ++j)
    {
      djv_image::pixel::F32_T * p =
        reinterpret_cast<djv_image::pixel::F32_T *>(out->data(i, j));

      p[1] /= sum;
    }
  }
}

}
}
}
