/*------------------------------------------------------------------------------
 lib/djv_image/io_pic_load.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include <djv_image/io_pic.h>

namespace djv_image {
namespace io_pic {

Load::Load() :
  _type(TYPE(0))
{
  _io.endian(mem::endian() != mem::MSB);
  
  _compress[0] = false;
  _compress[1] = false;
}

io::Plugin * Load::copy() const { return new Load; }

const Str & Load::name() const { return io_pic::name; }

const List<Str> & Load::extension_list() const
  { return io_pic::extension_list; }

void Load::open(const File & in, Info * info, Tag * tag, Time * time)
  throw (Error)
{
  DJV_DBG_OBJ("Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(_file.get(_file.seq.start()), info, tag);

  if (file::SEQ == _file.type)
    time->list = _file.seq.list;
}

void Load::open(const File & in, List<Info> * info, Tag * tag, Time * time)
  throw (Error)
{
  io::Load::open(in, info, tag, time);
}

namespace {
  
struct Header
{
  uint32_t magic;
  float version;
  char comment [80];
  char id [4];
  uint16_t width, height;
  float ratio;
  uint16_t fields;
  uint8_t pad [2];
};

enum CHANNEL
{
  CHANNEL_A = 0x10,
  CHANNEL_B = 0x20,
  CHANNEL_G = 0x40,
  CHANNEL_R = 0x80
};
  
struct Channel
{
  uint8_t chained;
  uint8_t size;
  uint8_t type;
  uint8_t channel;
};

void _channel(djv_base::Io * in, Channel * out)
{
  in->get_u8(&out->chained);
  in->get_u8(&out->size);
  in->get_u8(&out->type);
  in->get_u8(&out->channel);
}

Str dbg_channel(int in)
{
  Str out;
  if (in & CHANNEL_R) out += 'R';
  if (in & CHANNEL_G) out += 'G';
  if (in & CHANNEL_B) out += 'B';
  if (in & CHANNEL_A) out += 'A';
  return out;
}

}

void Load::_open(const File & in, Info * info, Tag * tag) throw (Error)
{
  DJV_DBG_OBJ("Load::_open");
  DJV_DBG_PRN("in = " << in);
  
  // Open.
  
  _io.open(in, djv_base::io::READ);
  
  // Header.
  
  Header header;
  mem::zero(&header, sizeof(Header));
  
  _io.get_u32(&header.magic);
  if (header.magic != 0x5380F634)
    throw Error(name(), io::error_unrecognized);

  _io.get_f32(&header.version);
  _io.get(header.comment, sizeof(header.comment));
  _io.get(header.id, sizeof(header.id));
  _io.get_u16(&header.width);
  _io.get_u16(&header.height);
  _io.get_f32(&header.ratio);
  _io.get_u16(&header.fields);
  _io.seek(sizeof(header.pad));
  DJV_DBG_PRN("version = " << header.version);
  DJV_DBG_PRN("comment = " << Str(header.comment, sizeof(header.comment)));
  DJV_DBG_PRN("id = " << Str(header.id, sizeof(header.id)));
  DJV_DBG_PRN("width = " << header.width);
  DJV_DBG_PRN("height = " << header.height);
  DJV_DBG_PRN("ratio = " << header.ratio);
  DJV_DBG_PRN("fields = " << header.fields);

  if (Str(header.id, sizeof(header.id)) != "PICT")
    throw Error(name(), io::error_unsupported);
  
  // Information.
  
  info->name = in;
  
  info->size = Vec2i(header.width, header.height);
  
  info->mirror.y = true;

  Channel channel;
  mem::zero(&channel, sizeof(Channel));
  _channel(&_io, &channel);
  _compress[0] = 2 == channel.type;
  DJV_DBG_PRN("channel = " << dbg_channel(channel.channel));
  
  int type = -1;
  if (
    (CHANNEL_R & channel.channel) &&
    (CHANNEL_G & channel.channel) &&
    (CHANNEL_B & channel.channel) &&
    (CHANNEL_A & channel.channel) &&
    8 == channel.size &&
    ! channel.chained
  ) {
    type = TYPE_RGBA;
  }
  else if (
    (CHANNEL_R & channel.channel) &&
    (CHANNEL_G & channel.channel) &&
    (CHANNEL_B & channel.channel) &&
    8 == channel.size &&
    ! channel.chained
  ) {
    type = TYPE_RGB;
  }
  else if (
    (CHANNEL_R & channel.channel) &&
    (CHANNEL_G & channel.channel) &&
    (CHANNEL_B & channel.channel) &&
    8 == channel.size &&
    channel.chained
  ) {
    mem::zero(&channel, sizeof(Channel));
    _channel(&_io, &channel);
    _compress[1] = 2 == channel.type;
    DJV_DBG_PRN("channel = " << dbg_channel(channel.channel));
    
    if (
      ! (CHANNEL_R & channel.channel) &&
      ! (CHANNEL_G & channel.channel) &&
      ! (CHANNEL_B & channel.channel) &&
        (CHANNEL_A & channel.channel) &&
      8 == channel.size &&
      ! channel.chained
    ) {
      type = TYPE_RGB_A;
    }
  }
  if (-1 == type)
    throw Error(name(), io::error_unsupported);

  _type = TYPE(type);
  switch (_type)
  {
    case TYPE_RGB: info->pixel = pixel::RGB_U8; break;
    case TYPE_RGBA:
    case TYPE_RGB_A: info->pixel = pixel::RGBA_U8; break;
  }
  
  // Image tags.
  
  if (header.comment[0])
    (*tag)[tag::label_tag()[tag::DESCRIPTION]] =
      Str(header.comment, sizeof(header.comment));

  DJV_DBG_PRN("info = " << *info);
}

const Image * Load::get(int frame, int, data::PROXY proxy) throw (Error)
{
  DJV_DBG_OBJ("Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("proxy = " << proxy);
  
  // Open.
  
  Info info;
  _open(_file.get(frame != -1 ? frame : _file.seq.start()), &info, &_image.tag);

  // Read.
  
  _io.readahead();
  
  Data * data = proxy ? &_tmp : &_image;
  data->set(info);
  
  const int channels = pixel::channels(info.pixel);
  const int bytes = pixel::channel_bytes(info.pixel);
  const bool endian = _io.endian();
  
  const uint8_t * p = _io.mmap_p(), * const end = _io.mmap_end();
  for (int y = 0; y < info.size.y; ++y)
  {
    DJV_DBG_PRN("y = " << y);
    
    switch (_type)
    {
      case TYPE_RGB:
      case TYPE_RGBA:
        if (_compress[0])
        {
          p = rle_load(p, end, data->data(0, y), info.size.x, channels,
            channels * bytes, endian);
          if (! p)
            throw Error(name(), io::error_read);
        }
        else
        {
          const int size = info.size.x * channels * bytes;
          mem::copy(p, data->data(0, y), size);
          p += size;
        }
        break;

      case TYPE_RGB_A:
        if (_compress[0])
        {
          p = rle_load(p, end, data->data(0, y), info.size.x, 3,
            channels * bytes, endian);
          if (! p)
            throw Error(name(), io::error_read);
        }
        else
        {
          const int size = info.size.x * 3 * bytes;
          mem::copy(p, data->data(0, y), size);
          p += size;
        }
        if (_compress[1])
        {
          p = rle_load(p, end, data->data(0, y) + 3 * bytes, info.size.x, 1,
            channels * bytes, endian);
          if (! p)
            throw Error(name(), io::error_read);
        }
        else
        {
          const int size = info.size.x * 1 * bytes;
          mem::copy(p, data->data(0, y), size);
          p += size;
        }
        break;
    }
  }

  if (proxy)
  {
    info.size = data::proxy_scale(info.size, proxy);
    info.proxy = proxy;
    _image.set(info);
    data::proxy_scale(_tmp, &_image, proxy);
  }
  
  DJV_DBG_PRN("image = " << _image);
  return &_image;
}

}
}
