/*------------------------------------------------------------------------------
 lib/djv_image/io_sgi_info.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include <djv_image/io_sgi.h>

namespace djv_image {
namespace io_sgi {

namespace {

struct Header
{
  Header();

  void load(djv_base::io::Io *, Info *, bool * compress) throw (Error);
  void save(djv_base::io::Io *, const Info &, bool compress) throw (Error);
  
  void dbg() const;

private:

  struct Data
  {
    uint16_t magic;
    uint8_t  storage;
    uint8_t  bytes;
    uint16_t dimension;
    uint16_t width;
    uint16_t height;
    uint16_t channels;
    uint32_t pixel_min;
    uint32_t pixel_max;
  } _data;
};

Header::Header()
{
  _data.magic = 474;
  _data.storage = 0;
  _data.bytes = 0;
  _data.dimension = 0;
  _data.width = 0;
  _data.height = 0;
  _data.channels = 0;
  _data.pixel_min = 0;
  _data.pixel_max = 0;
}

void Header::load(djv_base::io::Io * io, Info * info, bool * compress)
  throw (Error)
{
  DJV_DBG_OBJ("Header::load");
  
  // Read.
  
  io->get_u16(&_data.magic);
  if (_data.magic != 474)
    throw Error(name, io::error_unrecognized);
  
  io->get_u8(&_data.storage);
  io->get_u8(&_data.bytes);
  io->get_u16(&_data.dimension);
  io->get_u16(&_data.width);
  io->get_u16(&_data.height);
  io->get_u16(&_data.channels);
  io->get_u32(&_data.pixel_min);
  io->get_u32(&_data.pixel_max);
  
  io->position(512);
  
  // Information.

  info->size = Vec2i(_data.width, _data.height);

  if (! pixel::pixel_(_data.channels, 1 == _data.bytes ? 8 : 16, false,
    &info->pixel)
  )
    throw Error(name, io::error_unsupported);
  
  info->endian = mem::MSB;

  *compress = _data.storage ? true : false;
}

void Header::save(djv_base::io::Io * io, const Info & info, bool compress)
  throw (Error)
{
  DJV_DBG_OBJ("Header::save");
  
  // Information.
  
  const int channels = pixel::channels(info.pixel);
  const int bytes = pixel::channel_bytes(info.pixel);
  
  _data.width = info.size.x;
  _data.height = info.size.y;

  _data.channels = channels;
  _data.bytes = bytes;

  _data.dimension = 1 == channels ? (1 == info.size.y ? 1 : 2) : 3;

  _data.pixel_min = 0;
  _data.pixel_max = 1 == bytes ? 255 : 65535;

  _data.storage = compress;
  
  // Write.
  
  io->set_u16(_data.magic);
  io->set_u8(_data.storage);
  io->set_u8(_data.bytes);
  io->set_u16(_data.dimension);
  io->set_u16(_data.width);
  io->set_u16(_data.height);
  io->set_u16(_data.channels);
  io->set_u32(_data.pixel_min);
  io->set_u32(_data.pixel_max);
  
  io->position(512);
}

void Header::dbg() const
{
  DJV_DBG_OBJ("Header::dbg");
  DJV_DBG_PRN("magic = " << _data.magic);
  DJV_DBG_PRN("storage = " << _data.storage);
  DJV_DBG_PRN("bytes = " << _data.bytes);
  DJV_DBG_PRN("dimension = " << _data.dimension);
  DJV_DBG_PRN("width = " << _data.width);
  DJV_DBG_PRN("height = " << _data.height);
  DJV_DBG_PRN("pixel range = " << _data.pixel_min << " " << _data.pixel_max);
}

}

void info_load(djv_base::io::Io * io, Info * info, bool * compress)
  throw (Error)
{
  Header header;
  header.load(io, info, compress);
}

void info_save(djv_base::io::Io * io, const Info & info, bool compress)
  throw (Error)
{
  Header header;
  header.save(io, info, compress);
}

}
}

