/*------------------------------------------------------------------------------
 plugin/djv_jpeg/save.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "io.h"

namespace djv_jpeg {
namespace io {
using libjpeg::jpeg_compress_struct;

extern "C" {
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save() { return new Save; }
}

Save::Save() :
  _f(0),
  _jpeg_init(false)
{}

Save::~Save()
{
  _close();
}

plugin::Plugin * Save::copy() const
{
  Save * out = new Save;
  out->_option = _option;
  return out;
}

const Str & Save::name() const { return djv_jpeg::name; }

const List<Str> & Save::extension_list() const
  { return djv_jpeg::extension_list; }

bool Save::option(const Str & in, Str * data)
{
try
{
  if (str::compare(in, option_list()[SAVE_QUALITY])) *data >> _option.quality;
}
catch (Str) { return false; }
  return true;
}

Str Save::option(const Str & in) const
{
  Str out;
  if (str::compare(in, option_list()[SAVE_QUALITY])) out << _option.quality;
  return out;
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const Info & info, const Tag &,
  const djv_image::Time & time) throw (Error)
{
  DJV_DBG_OBJ("Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  if (time.list.size() > 1) _file.type = file::SEQ;
  
  _info = Info();
  _info.size = info.size;
  pixel::FORMAT format = pixel::format(info.pixel);
  switch (format)
  {
    case pixel::LA: format = pixel::L; break;
    case pixel::RGBA: format = pixel::RGB; break;
    default: break;
  }
  _info.pixel = pixel::pixel(format, pixel::U8);
  DJV_DBG_PRN("info = " << _info);
  
  _image.set(_info);
}

namespace {

bool init(jpeg_compress_struct * jpeg, Jpeg_Error * error)
{
  if (setjmp(error->jump)) return false;
  libjpeg::jpeg_create_compress(jpeg);
  return true;
}

bool open(FILE * f, jpeg_compress_struct * jpeg, const Info & info,
  const Tag & tag, int quality, Jpeg_Error * error)
{
  if (setjmp(error->jump)) return false;

  libjpeg::jpeg_stdio_dest(jpeg, f);
  
  jpeg->image_width = info.size.x;
  jpeg->image_height = info.size.y;
  
  if (pixel::L_U8 == info.pixel)
  {
    jpeg->input_components = 1;
    jpeg->in_color_space = libjpeg::JCS_GRAYSCALE;
  }
  else if (pixel::RGB_U8 == info.pixel)
  {
    jpeg->input_components = 3;
    jpeg->in_color_space = libjpeg::JCS_RGB;
  }
  
  libjpeg::jpeg_set_defaults(jpeg);
  libjpeg::jpeg_set_quality(jpeg, quality, TRUE);
  libjpeg::jpeg_start_compress(jpeg, TRUE);
  
  Str _tag = tag[tag::label_tag()[tag::DESCRIPTION]];
  if (_tag.size())
    libjpeg::jpeg_write_marker(
      jpeg, JPEG_COM, (libjpeg::JOCTET *)_tag.c_str(), _tag.size());
  
  return true;
}

}

void Save::_open(const File & in, const Tag & tag) throw (Error)
{
  DJV_DBG_OBJ("Save::_open");
  DJV_DBG_PRN("in = " << in);
  
  _close();
  
  // Initialize libjpeg.
  
  _jpeg.err = libjpeg::jpeg_std_error(&_jpeg_error.pub);
  _jpeg_error.pub.error_exit = djv_jpeg_error;
  _jpeg_error.pub.emit_message = djv_jpeg_warn;
  _jpeg_error.msg[0] = 0;
  
  if (! io::init(&_jpeg, &_jpeg_error))
    throw Error(name(), _jpeg_error.msg);
  
  _jpeg_init = true;
  
  // Open.
  
  _f = ::fopen(in.get().c_str(), "wb");
  if (! _f)
    throw Error(name(), djv_image::io::error_open);
  
  if (! io::open(_f, &_jpeg, _info, tag, _option.quality, &_jpeg_error))
    throw Error(name(), _jpeg_error.msg);
}

void Save::_close()
{
  if (_jpeg_init)
  {
    libjpeg::jpeg_destroy_compress(&_jpeg);
    _jpeg_init = false;
  }
  if (_f) { ::fclose(_f); _f = 0; }
}

namespace {

bool scanline(libjpeg::jpeg_compress_struct * jpeg, const uint8_t * in,
  Jpeg_Error * error)
{
  if (::setjmp(error->jump)) return false;
  libjpeg::JSAMPROW p [] = { (libjpeg::JSAMPLE *)in };
  if (! libjpeg::jpeg_write_scanlines(jpeg, p, 1)) return false;
  return true;
}

bool end(libjpeg::jpeg_compress_struct * jpeg, Jpeg_Error * error)
{
  if (::setjmp(error->jump)) return false;
  libjpeg::jpeg_finish_compress(jpeg);
  return true;
}

}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ("Save::set");
  DJV_DBG_PRN("in = " << in);
  
  // Open.

  _open(_file.get(frame), in.tag);
  
  // Write.
  
  const Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl_data::copy(in, &_image);
    p = &_image;
  }
  
  const int h = p->h();
  for (int y = 0; y < h; ++y)
    if (! scanline(&_jpeg, p->data(0, h - 1 - y), &_jpeg_error))
      throw Error(name(), _jpeg_error.msg);
  
  if (! end(&_jpeg, &_jpeg_error))
    throw Error(name(), _jpeg_error.msg);

  _close();
}

const List<Str> & label_save_option()
{
  static const List<Str> data = List<Str>() <<
    "Quality";
  return data;
}

}
}

