/*------------------------------------------------------------------------------
 plugin/djv_tiff/load.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "io.h"

namespace djv_tiff {
namespace io {

extern "C" {

DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_load()
{
  tiff_init();

  return new Load;
}

}

Load::Load() :
  _f(0)
{}

Load::~Load()
{
  _close();
}

plugin::Plugin * Load::copy() const { return new Load; }

const Str & Load::name() const { return djv_tiff::name; }

const List<Str> & Load::extension_list() const
  { return djv_tiff::extension_list; }

void Load::open(const File & in, Info * info, Tag * tag,
  djv_image::Time * time) throw (Error)
{
  DJV_DBG_OBJ("Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(_file.get(_file.seq.start()), info, tag);

  if (file::SEQ == _file.type)
    time->list = _file.seq.list;
}

void Load::open(const File & in, List<Info> * info, Tag * tag,
  djv_image::Time * time) throw (Error)
{
  djv_image::io::Load::open(in, info, tag, time);
}

void Load::_open(const File & in, Info * info, Tag * tag)
  throw (Error)
{
  DJV_DBG_OBJ("Load::_open");
  DJV_DBG_PRN("in = " << in);

  _close();
  
  // Open.
  
  _f = TIFFOpen(in.get().c_str(), "r");
  if (! _f)
    throw Error(name(), djv_image::io::error_open);
  
  // Header.

  uint32 width, height;
  uint16 photometric, samples, sample_depth, sample_format;
  uint16 * extra_samples, extra_samples_size;
  uint16 orient, compress, channels;
  
  TIFFGetFieldDefaulted(_f, TIFFTAG_IMAGEWIDTH, &width);
  TIFFGetFieldDefaulted(_f, TIFFTAG_IMAGELENGTH, &height);
  TIFFGetFieldDefaulted(_f, TIFFTAG_PHOTOMETRIC, &photometric);
  TIFFGetFieldDefaulted(_f, TIFFTAG_SAMPLESPERPIXEL, &samples);
  TIFFGetFieldDefaulted(_f, TIFFTAG_BITSPERSAMPLE, &sample_depth);
  TIFFGetFieldDefaulted(_f, TIFFTAG_SAMPLEFORMAT, &sample_format);
  TIFFGetFieldDefaulted(_f, TIFFTAG_EXTRASAMPLES, &extra_samples_size,
    &extra_samples);
  TIFFGetFieldDefaulted(_f, TIFFTAG_ORIENTATION, &orient);
  TIFFGetFieldDefaulted(_f, TIFFTAG_COMPRESSION, &compress);
  TIFFGetFieldDefaulted(_f, TIFFTAG_PLANARCONFIG, &channels);
  TIFFGetFieldDefaulted(_f, TIFFTAG_COLORMAP,
    &_colormap[0], &_colormap[1], &_colormap[2]);
  
  // Information.
  
  info->name = in;
  
  info->size = Vec2i(width, height);
  
  if (PLANARCONFIG_SEPARATE == channels)
    throw Error(name(), djv_image::io::error_unsupported);

  int pixel = -1;
  switch (photometric)
  {
    case PHOTOMETRIC_PALETTE: pixel = pixel::RGB_U8; break;
    case PHOTOMETRIC_MINISWHITE:
    case PHOTOMETRIC_MINISBLACK:
    case PHOTOMETRIC_RGB:
      if (32 == sample_depth && sample_format != SAMPLEFORMAT_IEEEFP) break;
      pixel = pixel::pixel_(samples, sample_depth, true);
      break;
  }
  if (-1 == pixel)
    throw Error(name(), djv_image::io::error_unsupported);
  info->pixel = PIXEL(pixel);

  _compress = compress != COMPRESSION_NONE;
  _palette = PHOTOMETRIC_PALETTE == photometric;
  
  switch (orient)
  {
    case ORIENTATION_TOPLEFT: info->mirror.y = true; break;
    case ORIENTATION_TOPRIGHT: info->mirror.x = info->mirror.y = true; break;
    case ORIENTATION_BOTRIGHT: info->mirror.x = true; break;
    case ORIENTATION_BOTLEFT: break;
  }
  
  // Image tags.
  
  const List<Str> & label_tag = tag::label_tag();
  char * _tag;
  if (TIFFGetField(_f, TIFFTAG_ARTIST, &_tag))
    (*tag)[label_tag[tag::CREATOR]] = _tag;
  if (TIFFGetField(_f, TIFFTAG_IMAGEDESCRIPTION, &_tag))
    (*tag)[label_tag[tag::DESCRIPTION]] = _tag;
  if (TIFFGetField(_f, TIFFTAG_COPYRIGHT, &_tag))
    (*tag)[label_tag[tag::COPYRIGHT]] = _tag;
  if (TIFFGetField(_f, TIFFTAG_DATETIME, &_tag))
    (*tag)[label_tag[tag::TIME]] = _tag;
}

void Load::_close()
{
  if (_f) { TIFFClose(_f); _f = 0; }
}

const Image * Load::get(int frame, int, data::PROXY proxy) throw (Error)
{
  DJV_DBG_OBJ("Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("proxy = " << proxy);
  
  // Open.
  
  Info info;
  _open(_file.get(frame != -1 ? frame : _file.seq.start()), &info, &_image.tag);
  
  // Read.
  
  Data * data = proxy ? &_tmp : &_image;
  data->set(info);
  
  for (int y = 0; y < info.size.y; ++y)
  {
    if (TIFFReadScanline(_f, (tdata_t *)data->data(0, y), y) == -1)
      throw Error(name(), djv_image::io::error_read);

    if (_palette)
      palette_load(data->data(0, y), info.size.x,
        pixel::channel_bytes(info.pixel),
        _colormap[0], _colormap[1], _colormap[2]);
  }

  if (proxy)
  {
    info.size = data::proxy_scale(info.size, proxy);
    info.proxy = proxy;
    _image.set(info);
    data::proxy_scale(_tmp, &_image, proxy);
  }
  
  DJV_DBG_PRN("image = " << _image);
  return &_image;
}

}
}

